"""This file contains utility format strings which are templated into many of our prompts."""

### FORMATTING INSTRUCTIONS ###
ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION = "Remember, your answer must be a valid JSON dictionary. The dictionary should only have the specified output fields."
ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION = "Remember, your answer must be a valid JSON list of dictionaries. The list may contain one or more dictionaries, and each dictionary should only have the specified output fields."

### USER-PROVIDED DESCRIPTION FOR MAPS / FILTERS / JOINS ###
DESC_SECTION = """
The user has additionally provided you with this description of the task you need to perform:
{desc}
"""

### JOB INSTRUCTIONS ###
AGG_JOB_INSTRUCTION = """analyze input {modalities} in order to perform an aggregation and generate a JSON object"""
MAP_JOB_INSTRUCTION = """analyze input {modalities} in order to produce a JSON object"""
FILTER_JOB_INSTRUCTION = """analyze input {modalities} in order to answer a TRUE / FALSE question"""
JOIN_JOB_INSTRUCTION = """analyze input {modalities} in order to determine whether two data records satisfy a join condition"""
PROPOSER_JOB_INSTRUCTION = """analyze input {modalities} in order to produce an answer to a question"""

### AGG / FILTER / JOIN CONDITIONS ###
EXAMPLE_AGG_INSTRUCTION = "Count the distinct number of scientists in the input."
EXAMPLE_FILTER_CONDITION = "The subject of the input is a foundational computer scientist."
EXAMPLE_JOIN_CONDITION = "The two inputs are scientists in the same academic field."

### EXAMPLE INPUT FIELDS ###
TEXT_EXAMPLE_INPUT_FIELDS = """
- text: a text passage describing a scientist
- birthday: the scientist's birthday
"""
IMAGE_EXAMPLE_INPUT_FIELDS = """
- image: an image of the scientist
- photographer: the photographer of the image
"""
AUDIO_EXAMPLE_INPUT_FIELDS = """
- recording: an audio recording of a newscast about the scientist's contributions to their field
- speaker: the speaker in the recording
"""
RIGHT_TEXT_EXAMPLE_INPUT_FIELDS = """
- contents: the contents of a text file
"""
RIGHT_IMAGE_EXAMPLE_INPUT_FIELDS = """
- headshot: a headshot of a famous scientist
"""
RIGHT_AUDIO_EXAMPLE_INPUT_FIELDS = """
- podcast: an audio recording of a podcast about historic scientists
"""

### EXAMPLE OUTPUT FIELDS ###
TEXT_EXAMPLE_OUTPUT_FIELDS = """- name: the name of the scientist
- birth_year: the year the scientist was born"""
IMAGE_EXAMPLE_OUTPUT_FIELDS = """- is_bald: true if the scientist is bald and false otherwise"""
AUDIO_EXAMPLE_OUTPUT_FIELDS = """- birthplace: the city where the scientist was born"""
AGG_EXAMPLE_OUTPUT_FIELDS = """- num_distinct_scientists: the number of distinct scientists mentioned in the input"""

### EXAMPLE CONTEXTS ###
TEXT_EXAMPLE_CONTEXT = """
  "text": "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace, was an English mathematician and writer chiefly known for her work on Charles Babbage's proposed mechanical general-purpose computer, the Analytical Engine. She was the first to recognise that the machine had applications beyond pure calculation.",
  "birthday": "December 10, 1815"
"""
IMAGE_EXAMPLE_CONTEXT = """
  "image": <bytes>,
  "photographer": "CameraEnthusiast1"
"""
AUDIO_EXAMPLE_CONTEXT = """
  "recording": <bytes>,
  "speaker": "Walter Cronkite"
"""
RIGHT_TEXT_EXAMPLE_CONTEXT = """
  "content": "Alan Turing was a pioneering computer scientist and mathematician. He is widely considered to be the father of theoretical computer science and artificial intelligence."
"""
RIGHT_IMAGE_EXAMPLE_CONTEXT = """
  "headshot": <bytes>
"""
RIGHT_AUDIO_EXAMPLE_CONTEXT = """
  "podcast": <bytes>
"""
SECOND_TEXT_EXAMPLE_CONTEXT = """
  "text": "Alan Turing was a pioneering computer scientist and mathematician. He is widely considered to be the father of theoretical computer science and artificial intelligence.",
  "birthday": "June 23, 1912"
"""
SECOND_IMAGE_EXAMPLE_CONTEXT = """
  "image": <bytes>,
  "photographer": "PhotoPro42"
"""
SECOND_AUDIO_EXAMPLE_CONTEXT = """
  "recording": <bytes>,
  "speaker": "Barbara Walters"
"""
THIRD_TEXT_EXAMPLE_CONTEXT = """
  "text": "Ada Lovelace is a historically significant computer scientist.",
  "birthday": "December 10, 1815"
"""
THIRD_IMAGE_EXAMPLE_CONTEXT = """
  "image": <bytes>,
  "photographer": "PicturePerfect"
"""
THIRD_AUDIO_EXAMPLE_CONTEXT = """
  "recording": <bytes>,
  "speaker": "Anderson Cooper"
"""

### DISCLAIMERS ###
IMAGE_DISCLAIMER = """
\n<image content provided here; assume in this example the image shows Ada Lovelace wearing a hat on top of her hair>
"""
AUDIO_DISCLAIMER = """
\n<audio content provided here; assume in this example the recording is about Ada Lovelace's upbringing in London>
"""
RIGHT_IMAGE_DISCLAIMER = """
\n<image content provided here; assume in this example the image shows Alan Turing working at his desk>
"""
RIGHT_AUDIO_DISCLAIMER = """
\n<audio content provided here; assume in this example the podcast is discussing Alan Turing's work on the Enigma code>
"""
AGG_IMAGE_DISCLAIMER = """
\n<image content provided here; assume in this example the first image shows Ada Lovelace, the second image shows Alan Turing, and the third image shows Ada Lovelace again>
"""
AGG_AUDIO_DISCLAIMER = """
\n<audio content provided here; assume in this example the first recording is about Ada Lovelace, the second recording is about Alan Turing, and the third recording is about Ada Lovelace again>
"""

### EXAMPLE REASONINGS ###
TEXT_EXAMPLE_REASONING = """The text passage mentions the scientist's name as "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace" and the scientist's birthday as "December 10, 1815". Therefore, the name of the scientist is "Augusta Ada King" and the birth year is 1815."""
IMAGE_EXAMPLE_REASONING = """The image shows hair on top of the scientist's head, so the is_bald field should be false."""
AUDIO_EXAMPLE_REASONING = """The newscast recording discusses Ada Lovelace's upbringing in London, so the birthplace field should be "London"."""
AGG_EXAMPLE_REASONING = """The input contains two distinct scientists: "Augusta Ada King" and "Alan Turing". Although "Ada Lovelace" is mentioned twice, she should only be counted once. Therefore, the number of distinct scientists mentioned in the input is 2."""
FILTER_EXAMPLE_REASONING = """Ada Lovelace is a foundational computer scientist, therefore the answer is TRUE."""
JOIN_EXAMPLE_REASONING = """The subject of the left record is Ada Lovelace and the subject of the right record is Alan Turing. Since both inputs are about computer scientists, they satisfy the join condition. Therefore, the answer is TRUE."""

### EXAMPLE ANSWERS ###
AGG_EXAMPLE_ANSWER = """
  "num_distinct_scientists": 2
"""
TEXT_EXAMPLE_ANSWER = """
  "name": "Augusta Ada King",
  "birth_year": 1815
"""
IMAGE_EXAMPLE_ANSWER = """
  "is_bald": false,
"""
AUDIO_EXAMPLE_ANSWER = """
  "birthplace": "London",
"""
TEXT_SENTENCE_EXAMPLE_ANSWER = """the text passage mentions the scientist's name as "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace" and the scientist's birthday as "December 10, 1815". Therefore, the name of the scientist is "Augusta Ada King" and the birth year is 1815."""
IMAGE_SENTENCE_EXAMPLE_ANSWER = """The image shows hair on top of the woman's head, so the is_bald field should be false."""
AUDIO_SENTENCE_EXAMPLE_ANSWER = """The newscast recording discusses Ada Lovelace's upbringing in London, so her birthplace is "London"."""



import logging
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed

import numpy as np
from chromadb.api.models.Collection import Collection

from palimpzest.constants import Cardinality
from palimpzest.core.data.dataset import Dataset
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.models import GenerationStats, PlanStats, SentinelPlanStats
from palimpzest.policy import Policy
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.filter import LLMFilter
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ContextScanOp, ScanPhysicalOp
from palimpzest.query.operators.topk import TopKOp
from palimpzest.query.optimizer.plan import PhysicalPlan, SentinelPlan
from palimpzest.utils.progress import PZSentinelProgressManager
from palimpzest.validator.validator import Validator

logger = logging.getLogger(__name__)

class BaseExecutionStrategy:
    def __init__(self,
                 scan_start_idx: int = 0, 
                 max_workers: int | None = None,
                 batch_size: int | None = None,
                 num_samples: int | None = None,
                 verbose: bool = False,
                 progress: bool = True,
                 *args,
                 **kwargs):
        self.scan_start_idx = scan_start_idx
        self.max_workers = max_workers
        self.batch_size = batch_size
        self.num_samples = num_samples
        self.verbose = verbose
        self.progress = progress


class ExecutionStrategy(BaseExecutionStrategy, ABC):
    """Base strategy for executing query plans. Defines how to execute a PhysicalPlan.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.info(f"Initialized ExecutionStrategy {self.__class__.__name__}")
        logger.debug(f"ExecutionStrategy initialized with config: {self.__dict__}")

    @abstractmethod
    def execute_plan(self, plan: PhysicalPlan) -> tuple[list[DataRecord], PlanStats]:
        """Execute a single plan according to strategy"""
        pass

    def _create_input_queues(self, plan: PhysicalPlan) -> dict[str, dict[str, list]]:
        """Initialize input queues for each operator in the plan."""
        input_queues = {f"{topo_idx}-{op.get_full_op_id()}": {} for topo_idx, op in enumerate(plan)}
        for topo_idx, op in enumerate(plan):
            full_op_id = op.get_full_op_id()
            unique_op_id = f"{topo_idx}-{full_op_id}"
            if isinstance(op, ScanPhysicalOp):
                scan_end_idx = (
                    len(op.datasource)
                    if self.num_samples is None
                    else min(self.scan_start_idx + self.num_samples, len(op.datasource))
                )
                input_queues[unique_op_id][f"source_{full_op_id}"] = [idx for idx in range(self.scan_start_idx, scan_end_idx)]
            elif isinstance(op, ContextScanOp):
                input_queues[unique_op_id][f"source_{full_op_id}"] = [None]
            else:
                for source_unique_full_op_id in plan.get_source_unique_full_op_ids(topo_idx, op):
                    input_queues[unique_op_id][source_unique_full_op_id] = []

        return input_queues

class SentinelExecutionStrategy(BaseExecutionStrategy, ABC):
    """Base strategy for executing sentinel query plans. Defines how to execute a SentinelPlan."""
    """
    Specialized query processor that implements MAB sentinel strategy
    for coordinating optimization and execution.
    """
    def __init__(
        self,
        policy: Policy,
        k: int = 6,
        j: int = 4,
        sample_budget: int = 100,
        sample_cost_budget: float | None = None,
        priors: dict | None = None,
        use_final_op_quality: bool = False,
        seed: int = 42,
        exp_name: str | None = None,
        dont_use_priors: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.k = k
        self.j = j
        self.sample_budget = sample_budget
        self.sample_cost_budget = sample_cost_budget
        self.policy = policy
        self.priors = priors
        self.use_final_op_quality = use_final_op_quality
        self.seed = seed
        self.rng = np.random.default_rng(seed=seed)
        self.exp_name = exp_name
        self.dont_use_priors = dont_use_priors

        # general cache which maps hash(logical_op_id, phys_op_id, hash(input)) --> record_set
        self.cache: dict[int, DataRecordSet] = {}

        # progress manager used to track progress of the execution
        self.progress_manager: PZSentinelProgressManager | None = None

    def _score_quality(
        self,
        validator: Validator,
        source_indices_to_record_sets: dict[tuple[str], list[tuple[DataRecordSet, PhysicalOperator]]],
    ) -> tuple[dict[int, list[DataRecordSet]], GenerationStats]:
        # extract information about the logical operation performed at this stage of the sentinel plan;
        # NOTE: we can infer these fields from context clues, but in the long-term we should have a more
        #       principled way of getting these directly from attributes either stored in the sentinel_plan
        #       or in the PhysicalOperator
        def is_perfect_quality_op(op: PhysicalOperator):
            return (
                not isinstance(op, LLMConvert)
                and not isinstance(op, LLMFilter)
                and not isinstance(op, TopKOp)
                and not isinstance(op, JoinOp)
            )

        # create minimal set of futures necessary to compute quality of each output record
        futures, full_hashes, full_hash_to_bool_output = [], set(), {}
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            for _, record_set_tuples in source_indices_to_record_sets.items():
                for record_set, op in record_set_tuples:
                    # if this operation does not involve an LLM, every record_op_stats object gets perfect quality
                    if is_perfect_quality_op(op):
                        for record_op_stats in record_set.record_op_stats:
                            record_op_stats.quality = 1.0
                        continue

                    # if the operation failed, assign 0.0 quality
                    if len(record_set) == 0:
                        record_set.record_op_stats[0].quality = 0.0
                        continue

                    # create future for map
                    if isinstance(op, LLMConvert) and op.cardinality is Cardinality.ONE_TO_ONE:
                        fields = op.generated_fields
                        input_record: DataRecord = record_set.input
                        output = record_set.data_records[0].to_dict(project_cols=fields)
                        output_str = record_set.data_records[0].to_json_str(project_cols=fields, bytes_to_str=True, sorted=True)
                        full_hash = f"{hash(input_record)}{hash(output_str)}"
                        if full_hash not in full_hashes:
                            full_hashes.add(full_hash)
                            futures.append(executor.submit(validator._score_map, op, fields, input_record, output, full_hash))

                    # create future for flat map
                    elif isinstance(op, LLMConvert) and op.cardinality is Cardinality.ONE_TO_MANY:
                        fields = op.generated_fields
                        input_record: DataRecord = record_set.input
                        output, output_strs = [], []
                        for data_record in record_set.data_records:
                            output.append(data_record.to_dict(project_cols=fields))
                            output_strs.append(data_record.to_json_str(project_cols=fields, bytes_to_str=True, sorted=True))
                        full_hash = f"{hash(input_record)}{hash(tuple(sorted(output_strs)))}"
                        if full_hash not in full_hashes:
                            full_hashes.add(full_hash)
                            futures.append(executor.submit(validator._score_flat_map, op, fields, input_record, output, full_hash))

                    # create future for top-k
                    elif isinstance(op, TopKOp):
                        fields = op.generated_fields
                        input_record: DataRecord = record_set.input
                        output = record_set.data_records[0].to_dict(project_cols=fields)
                        output_str = record_set.data_records[0].to_json_str(project_cols=fields, bytes_to_str=True, sorted=True)
                        full_hash = f"{hash(input_record)}{hash(output_str)}"
                        if full_hash not in full_hashes:
                            full_hashes.add(full_hash)
                            futures.append(executor.submit(validator._score_topk, op, fields, input_record, output, full_hash))

                    # create future for filter
                    elif isinstance(op, LLMFilter):
                        filter_str = op.filter_obj.filter_condition
                        input_record: DataRecord = record_set.input
                        output = record_set.data_records[0]._passed_operator
                        full_hash = f"{filter_str}{hash(input_record)}"
                        if full_hash not in full_hashes:
                            full_hash_to_bool_output[full_hash] = output
                            full_hashes.add(full_hash)
                            futures.append(executor.submit(validator._score_filter, op, filter_str, input_record, output, full_hash))

                    # create future for join
                    elif isinstance(op, JoinOp):
                        condition = op.condition
                        for left_idx, left_input_record in enumerate(record_set.input[0]):
                            for right_idx, right_input_record in enumerate(record_set.input[1]):
                                record_idx = left_idx * len(record_set.input[1]) + right_idx
                                output = record_set.data_records[record_idx]._passed_operator
                                full_hash = f"{condition}{hash(left_input_record)}{hash(right_input_record)}"
                                if full_hash not in full_hashes:
                                    full_hash_to_bool_output[full_hash] = output
                                    full_hashes.add(full_hash)
                                    futures.append(executor.submit(validator._score_join, op, condition, left_input_record, right_input_record, output, full_hash))

        # collect results from futures
        full_hash_to_score, validation_gen_stats = {}, GenerationStats()
        for future in as_completed(futures):
            score, gen_stats, full_hash = future.result()
            full_hash_to_score[full_hash] = score
            validation_gen_stats += gen_stats

        # compute quality of each output computed by this operator
        for _, record_set_tuples in source_indices_to_record_sets.items():
            for record_set, op in record_set_tuples:
                if is_perfect_quality_op(op) or len(record_set) == 0:
                    continue

                if isinstance(op, LLMConvert) and op.cardinality is Cardinality.ONE_TO_ONE:
                    fields = op.generated_fields
                    input_record: DataRecord = record_set.input
                    output_str = record_set.data_records[0].to_json_str(project_cols=fields, bytes_to_str=True, sorted=True)
                    full_hash = f"{hash(input_record)}{hash(output_str)}"
                    record_set.record_op_stats[0].quality = full_hash_to_score[full_hash]

                elif isinstance(op, LLMConvert) and op.cardinality is Cardinality.ONE_TO_MANY:
                    fields = op.generated_fields
                    input_record: DataRecord = record_set.input
                    output_strs = []
                    for data_record in record_set.data_records:
                        output_strs.append(data_record.to_json_str(project_cols=fields, bytes_to_str=True, sorted=True))
                    full_hash = f"{hash(input_record)}{hash(tuple(sorted(output_strs)))}"
                    score = full_hash_to_score[full_hash]
                    for record_op_stats in record_set.record_op_stats:
                        record_op_stats.quality = score

                # TODO: this scoring function will (likely) bias towards small values of k since it
                # measures precision and not recall / F1; will need to revisit this in the future
                elif isinstance(op, TopKOp):
                    fields = op.generated_fields
                    input_record: DataRecord = record_set.input
                    output_str = record_set.data_records[0].to_json_str(project_cols=fields, bytes_to_str=True, sorted=True)
                    full_hash = f"{hash(input_record)}{hash(output_str)}"
                    score = full_hash_to_score[full_hash]
                    record_set.record_op_stats[0].quality = score

                elif isinstance(op, LLMFilter):
                    filter_str = op.filter_obj.filter_condition
                    input_record: DataRecord = record_set.input
                    output = record_set.data_records[0]._passed_operator
                    full_hash = f"{filter_str}{hash(input_record)}"
                    if output == full_hash_to_bool_output[full_hash]:
                        record_set.record_op_stats[0].quality = full_hash_to_score[full_hash]
                    else:
                        record_set.record_op_stats[0].quality = 1.0 - full_hash_to_score[full_hash]

                elif isinstance(op, JoinOp):
                    condition = op.condition
                    for left_idx, left_input_record in enumerate(record_set.input[0]):
                        for right_idx, right_input_record in enumerate(record_set.input[1]):
                            record_idx = left_idx * len(record_set.input[1]) + right_idx
                            output = record_set.data_records[record_idx]._passed_operator
                            full_hash = f"{condition}{hash(left_input_record)}{hash(right_input_record)}"
                            if output == full_hash_to_bool_output[full_hash]:
                                record_set.record_op_stats[record_idx].quality = full_hash_to_score[full_hash]
                            else:
                                record_set.record_op_stats[record_idx].quality = 1.0 - full_hash_to_score[full_hash]

        # return the quality annotated record sets
        return source_indices_to_record_sets, validation_gen_stats

    def _execute_op_set(self, unique_logical_op_id: str, op_inputs: list[tuple[PhysicalOperator, str | tuple, int | DataRecord | list[DataRecord] | tuple[list[DataRecord]]]]) -> tuple[dict[int, list[tuple[DataRecordSet, PhysicalOperator, bool]]], dict[str, int]]:
        def execute_op_wrapper(operator: PhysicalOperator, source_indices: str | tuple, input: int | DataRecord | list[DataRecord] | tuple[list[DataRecord]]) -> tuple[DataRecordSet, PhysicalOperator, list[DataRecord] | list[int]]:
            # operator is a join
            record_set = operator(input[0], input[1]) if isinstance(operator, JoinOp) else operator(input)
            return record_set, operator, source_indices, input

        def get_hash(operator: PhysicalOperator, input: int | DataRecord | list[DataRecord] | tuple[list[DataRecord]]):
            if isinstance(input, list):
                input = tuple(input)
            elif isinstance(input, tuple):
                input = (tuple(input[0]), tuple(input[1]))
            return hash(f"{operator.get_full_op_id()}{hash(input)}")

        # initialize mapping from source indices to output record sets
        source_indices_to_record_sets_and_ops = {source_indices: [] for _, source_indices, _ in op_inputs}

        # if any operations were previously executed, read the results from the cache
        final_op_inputs = []
        for operator, source_indices, input in op_inputs:
            # compute hash
            op_input_hash = get_hash(operator, input)

            # get result from cache
            if op_input_hash in self.cache:
                record_set, operator = self.cache[op_input_hash]
                source_indices_to_record_sets_and_ops[source_indices].append((record_set, operator, False))

            # otherwise, add to final_op_inputs
            else:
                final_op_inputs.append((operator, source_indices, input))

        # keep track of the number of llm operations
        num_llm_ops = 0

        # create thread pool w/max workers and run futures over worker pool
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            # create futures
            futures = [
                executor.submit(execute_op_wrapper, operator, source_indices, input)
                for operator, source_indices, input in final_op_inputs
            ]

            output_record_sets = []
            for future in as_completed(futures):
                # update output record sets
                record_set, operator, source_indices, input = future.result()

                # if the operator is a join, get record_set from tuple output
                if isinstance(operator, JoinOp):
                    record_set = record_set[0]

                output_record_sets.append((record_set, operator, source_indices, input))

                # update cache
                op_input_hash = get_hash(operator, input)
                self.cache[op_input_hash] = (record_set, operator)

                # update progress manager
                if self._is_llm_op(operator):
                    num_llm_ops += 1
                    self.progress_manager.incr(unique_logical_op_id, num_samples=1, total_cost=record_set.get_total_cost())

            # update mapping from source_indices to record sets and operators
            for record_set, operator, source_indices, input in output_record_sets:
                # add record_set to mapping from source_indices --> record_sets
                record_set.input = input
                source_indices_to_record_sets_and_ops[source_indices].append((record_set, operator, True))

        return source_indices_to_record_sets_and_ops, num_llm_ops

    def _is_llm_op(self, physical_op: PhysicalOperator) -> bool:
        is_llm_convert = isinstance(physical_op, LLMConvert)
        is_llm_filter = isinstance(physical_op, LLMFilter)
        is_llm_topk = isinstance(physical_op, TopKOp) and isinstance(physical_op.index, Collection)
        is_llm_join = isinstance(physical_op, JoinOp)
        return is_llm_convert or is_llm_filter or is_llm_topk or is_llm_join

    @abstractmethod
    def execute_sentinel_plan(self, sentinel_plan: SentinelPlan, train_dataset: dict[str, Dataset], validator: Validator) -> SentinelPlanStats:
        """Execute a SentinelPlan according to strategy"""
        pass



from __future__ import annotations

import logging
import warnings

import pandas as pd

from palimpzest.constants import NAIVE_BYTES_PER_RECORD
from palimpzest.core.models import OperatorCostEstimates, PlanCost, SentinelPlanStats
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ContextScanOp, MarshalAndScanDataOp, ScanPhysicalOp

warnings.simplefilter(action='ignore', category=UserWarning)

logger = logging.getLogger(__name__)

class BaseCostModel:
    """
    This base class contains the interface/abstraction that every CostModel must implement
    in order to work with the Optimizer. In brief, the Optimizer expects the CostModel to
    make a prediction about the runtime, cost, and quality of a physical operator.
    """
    def __init__(self):
        """
        CostModel constructor; the arguments for individual CostModels may vary depending
        on the assumptions they make about the prevalance of historical execution data
        and online vs. batch execution settings.
        """
        pass

    def get_costed_full_op_ids(self) -> set[str]:
        """
        Return the set of full op ids which the cost model has cost estimates for.
        """
        raise NotImplementedError("Calling get_costed_full_op_ids from abstract method")

    def __call__(self, operator: PhysicalOperator) -> PlanCost:
        """
        The interface exposed by the CostModel to the Optimizer. Subclasses may require
        additional arguments in order to make their predictions.
        """
        raise NotImplementedError("Calling __call__ from abstract method")


class SampleBasedCostModel:
    """
    """
    def __init__(
        self,
        sentinel_plan_stats: SentinelPlanStats | None = None,
        verbose: bool = False,
        exp_name: str | None = None,
    ):
        # store verbose argument
        self.verbose = verbose

        # store experiment name if one is provided
        self.exp_name = exp_name

        # construct cost, time, quality, and selectivity matrices for each operator set;
        self.operator_to_stats = self._compute_operator_stats(sentinel_plan_stats)
        self.costed_full_op_ids = None if self.operator_to_stats is None else set([
            full_op_id
            for _, full_op_id_to_stats in self.operator_to_stats.items()
            for full_op_id in full_op_id_to_stats
        ])

        # if there is a logical operator with no samples; add all of its op ids to costed_full_op_ids;
        # this will lead to the cost model applying the naive cost estimates for all physical op ids
        # in this logical operator (I think?)
        # TODO

        logger.info(f"Initialized SampleBasedCostModel with verbose={self.verbose}")
        logger.debug(f"Initialized SampleBasedCostModel with params: {self.__dict__}")

    def get_costed_full_op_ids(self):
        return self.costed_full_op_ids

    def _compute_operator_stats(self, sentinel_plan_stats: SentinelPlanStats | None) -> dict:
        logger.debug("Computing operator statistics")
        # if no stats are provided, simply return None
        if sentinel_plan_stats is None:
            return None

        # flatten the nested dictionary of execution data and pull out fields relevant to cost estimation
        execution_record_op_stats = []
        for unique_logical_op_id, full_op_id_to_op_stats in sentinel_plan_stats.operator_stats.items():
            logger.debug(f"Computing operator statistics for logical_op_id: {unique_logical_op_id}")
            # flatten the execution data into a list of RecordOpStats
            op_set_execution_data = [
                record_op_stats
                for _, op_stats in full_op_id_to_op_stats.items()
                for record_op_stats in op_stats.record_op_stats_lst
            ]

            # add entries from execution data into matrices
            for record_op_stats in op_set_execution_data:
                record_op_stats_dict = {
                    "unique_logical_op_id": unique_logical_op_id,
                    "full_op_id": record_op_stats.full_op_id,
                    "record_id": record_op_stats.record_id,
                    "record_parent_ids": record_op_stats.record_parent_ids,
                    "cost_per_record": record_op_stats.cost_per_record,
                    "time_per_record": record_op_stats.time_per_record,
                    "quality": record_op_stats.quality,
                    "passed_operator": record_op_stats.passed_operator,
                    "source_indices": record_op_stats.record_source_indices,
                    "op_details": record_op_stats.op_details,
                    "answer": record_op_stats.answer,
                    "op_name": record_op_stats.op_name,
                }
                execution_record_op_stats.append(record_op_stats_dict)

        # convert flattened execution data into dataframe
        operator_stats_df = pd.DataFrame(execution_record_op_stats)

        # for each full_op_id, compute its average cost_per_record, time_per_record, selectivity, and quality
        operator_to_stats = {}
        for unique_logical_op_id, logical_op_df in operator_stats_df.groupby("unique_logical_op_id"):
            logger.debug(f"Computing operator statistics for unique_logical_op_id: {unique_logical_op_id}")
            operator_to_stats[unique_logical_op_id] = {}

            for full_op_id, physical_op_df in logical_op_df.groupby("full_op_id"):
                # compute the number of input records processed by this operator; use source_indices for scan operator(s)
                num_source_records = (
                    physical_op_df.record_parent_ids.apply(tuple).nunique()
                    if not physical_op_df.record_parent_ids.isna().all()
                    else physical_op_df.source_indices.apply(tuple).nunique()
                )

                # compute selectivity; for filters this may be 1.0 on smalle samples;
                # always put something slightly less than 1.0 to ensure that filters are pushed down when possible
                selectivity = physical_op_df.passed_operator.sum() / num_source_records
                op_name = physical_op_df.op_name.iloc[0].lower()
                if selectivity == 1.0 and "filter" in op_name:
                    selectivity -= 1e-3

                # compute quality; if all qualities are None then this will be NaN
                quality = physical_op_df.quality.mean()

                # set operator stats for this physical operator
                operator_to_stats[unique_logical_op_id][full_op_id] = {
                    "cost": physical_op_df.cost_per_record.mean(),
                    "time": physical_op_df.time_per_record.mean(),
                    "quality": 1.0 if pd.isna(quality) else quality,
                    "selectivity": selectivity,
                }

        logger.debug(f"Done computing operator statistics for {len(operator_to_stats)} operators!")
        return operator_to_stats

    def _compute_naive_plan_cost(self, operator: PhysicalOperator, source_op_estimates: OperatorCostEstimates | None = None, right_source_op_estimates: OperatorCostEstimates | None = None) -> PlanCost:
        # get identifier for operator which is unique within sentinel plan but consistent across sentinels
        full_op_id = operator.get_full_op_id()
        logger.debug(f"Calling __call__ for {str(operator)} with full_op_id: {full_op_id}")

        # initialize estimates of operator metrics based on naive (but sometimes precise) logic
        if isinstance(operator, MarshalAndScanDataOp):
            # get handle to scan operator and pre-compute its size (number of records)
            datasource_len = len(operator.datasource)

            source_op_estimates = OperatorCostEstimates(
                cardinality=datasource_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

            op_estimates = operator.naive_cost_estimates(source_op_estimates, input_record_size_in_bytes=NAIVE_BYTES_PER_RECORD)

        elif isinstance(operator, ContextScanOp):
            source_op_estimates = OperatorCostEstimates(
                cardinality=1.0,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

            op_estimates = operator.naive_cost_estimates(source_op_estimates)

        elif isinstance(operator, JoinOp):
            op_estimates = operator.naive_cost_estimates(source_op_estimates, right_source_op_estimates)

        else:
            op_estimates = operator.naive_cost_estimates(source_op_estimates)

        # compute estimates for this operator
        est_input_cardinality = (
            source_op_estimates.cardinality * right_source_op_estimates.cardinality
            if isinstance(operator, JoinOp)
            else source_op_estimates.cardinality
        )
        op_time = op_estimates.time_per_record * est_input_cardinality
        op_cost = op_estimates.cost_per_record * est_input_cardinality
        op_quality = op_estimates.quality

        # create and return PlanCost object for this op's statistics
        op_plan_cost = PlanCost(
            cost=op_cost,
            time=op_time,
            quality=op_quality,
            op_estimates=op_estimates,
        )
        logger.debug(f"Done calling __call__ for {str(operator)} with full_op_id: {full_op_id}")
        logger.debug(f"Plan cost: {op_plan_cost}")

        return op_plan_cost

    def __call__(self, operator: PhysicalOperator, source_op_estimates: OperatorCostEstimates | None = None, right_source_op_estimates: OperatorCostEstimates | None = None) -> PlanCost:
        # for non-sentinel execution, we use naive estimates
        full_op_id = operator.get_full_op_id()
        unique_logical_op_id = operator.unique_logical_op_id
        if self.operator_to_stats is None or unique_logical_op_id not in self.operator_to_stats:
            return self._compute_naive_plan_cost(operator, source_op_estimates, right_source_op_estimates)

        # NOTE: some physical operators may not have any sample execution data in this cost model;
        #       these physical operators are filtered out of the Optimizer, thus we can assume that
        #       we will have execution data for each operator passed into __call__; nevertheless, we
        #       still perform a sanity check
        # look up physical and logical op ids associated with this physical operator
        physical_op_to_stats = self.operator_to_stats.get(unique_logical_op_id)
        assert physical_op_to_stats is not None, f"No execution data for logical operator: {str(operator)}"
        assert physical_op_to_stats.get(full_op_id) is not None, f"No execution data for physical operator: {str(operator)}"
        logger.debug(f"Calling __call__ for {str(operator)}")

        # look up stats for this operation
        est_cost_per_record = self.operator_to_stats[unique_logical_op_id][full_op_id]["cost"]
        est_time_per_record = self.operator_to_stats[unique_logical_op_id][full_op_id]["time"]
        est_quality = self.operator_to_stats[unique_logical_op_id][full_op_id]["quality"]
        est_selectivity = self.operator_to_stats[unique_logical_op_id][full_op_id]["selectivity"]

        # create source_op_estimates for scan operators if they are not provided
        if isinstance(operator, ScanPhysicalOp):
            # get handle to scan operator and pre-compute its size (number of records)
            datasource_len = len(operator.datasource)

            source_op_estimates = OperatorCostEstimates(
                cardinality=datasource_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

        # generate new set of OperatorCostEstimates
        est_input_cardinality = (
            source_op_estimates.cardinality * right_source_op_estimates.cardinality
            if isinstance(operator, JoinOp)
            else source_op_estimates.cardinality
        )
        op_estimates = OperatorCostEstimates(
            cardinality=est_selectivity * est_input_cardinality,
            time_per_record=est_time_per_record,
            cost_per_record=est_cost_per_record,
            quality=est_quality,
        )

        # compute estimates for this operator
        op_time = op_estimates.time_per_record * est_input_cardinality
        op_cost = op_estimates.cost_per_record * est_input_cardinality
        op_quality = op_estimates.quality

        # construct and return op estimates
        plan_cost = PlanCost(cost=op_cost, time=op_time, quality=op_quality, op_estimates=op_estimates)
        logger.debug(f"Done calling __call__ for {str(operator)}")
        logger.debug(f"Plan cost: {plan_cost}")
        return plan_cost



from __future__ import annotations

import warnings
from collections.abc import Iterator
from typing import Callable

from chromadb.api.models.Collection import Collection
from pydantic import BaseModel

from palimpzest.constants import AggFunc, Cardinality
from palimpzest.core.elements.filters import Filter
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.lib.schemas import create_schema_from_fields, project, relax_schema, union_schemas
from palimpzest.policy import construct_policy_from_kwargs
from palimpzest.query.operators.logical import (
    Aggregate,
    ConvertScan,
    Distinct,
    FilteredScan,
    GroupByAggregate,
    JoinOp,
    LimitScan,
    LogicalOperator,
    Project,
    TopKScan,
)
from palimpzest.query.processor.config import QueryProcessorConfig
from palimpzest.utils.hash_helpers import hash_for_serialized_dict
from palimpzest.validator.validator import Validator


# TODO?: remove `schema` from `Dataset` and access it from `operator`?
# - Q: how do you handle datasets with multiple sources?
#    - for joins the operator should have the union'ed schema
#    - but for Contexts it may be trickier
class Dataset:
    """
    A `Dataset` represents a collection of structured or unstructured data that can be processed and
    transformed. Each `Dataset` is either a "root" `Dataset` (which yields data items) or it is the
    result of performing data processing operation(s) on root `Dataset(s)`.

    Users can perform computations on a `Dataset` in a lazy or eager fashion. Applying functions
    such as `sem_filter`, `sem_map`, `sem_join`, `sem_agg`, etc. will lazily create a new `Dataset`.
    Users can invoke the `run()` method to execute the computation and retrieve a materialized `Dataset`.
    Materialized `Dataset`s can be processed further, or their results can be retrieved using `.get()`.
    
    A root `Dataset` must subclass at least one of `pz.IterDataset`, `pz.IndexDataset`, or `pz.Context`.
    Each of these classes supports a different access pattern:

        - `pz.IterDataset`: supports accessing data via iteration
            - Ex: iterating over a list of PDFs
            - Ex: iterating over rows in a DataFrame
        - `pz.IndexDataset`: supports accessing data via point lookups / queries
            - Ex: querying a vector database
            - Ex: querying a SQL database
        - `pz.Context`: supports accessing data with an agent
            - Ex: processing a set of CSV files with a data science agent
            - Ex: processing time series data with a data cleaning agent

    A root `Dataset` may subclass more than one of the aforementioned classes. For example, the root
    `Dataset` for a list of files may inherit from `pz.IterDataset` and `pz.IndexDataset` to support
    iterating over the files and performing point lookups for individual files.

    For details on how to create your own root `Dataset`, please see: TODO
    """
    def __init__(
            self,
            sources: list[Dataset] | Dataset | None,
            operator: LogicalOperator,
            schema: type[BaseModel] | None = None,
            id: str | None = None,
        ) -> None:
        """
        Initialize a `Dataset` with one or more `sources` and the operator that is being applied.
        Root `Datasets` subclass `pz.IterDataset`, `pz.IndexDataset`, and/or `pz.Context` and use
        their own constructors.

        Args:
            sources (`list[Dataset] | Dataset`): The (list of) `Dataset(s)` which are input(s) to
                the operator used to compute this `Dataset`.
            operator (`LogicalOperator`): The `LogicalOperator` used to compute this `Dataset`.
            schema (type[`BaseModel`] | None): The schema of this `Dataset`.
            id (str | None): an identifier for this `Dataset` provided by the user

        Raises:
            ValueError: if `sources` is not a `Dataset` or list of `Datasets`
        """
        # set sources
        self._sources = []
        if isinstance(sources, list):
            self._sources = sources
        elif isinstance(sources, Dataset):
            self._sources = [sources]
        elif sources is not None:
            raise ValueError("Dataset sources must be another Dataset or a list of Datasets. For root Datasets, you must subclass pz.IterDataset, pz.IndexDataset, or pz.Context.")

        # set the logical operator and schema
        self._operator: LogicalOperator = operator
        self._schema = schema

        # compute the dataset id
        self._id = self._compute_dataset_id() if id is None else id

    @property
    def id(self) -> str:
        """The string identifier for this `Dataset`"""
        return self._id

    @property
    def schema(self) -> type[BaseModel]:
        """The Pydantic model defining the schema of this `Dataset`"""
        return self._schema

    @property
    def is_root(self) -> bool:
        return len(self._sources) == 0

    def __str__(self) -> str:
        return f"Dataset(schema={self._schema}, id={self._id}, op_id={self._operator.get_logical_op_id()})"

    def __iter__(self) -> Iterator[Dataset]:
        for source in self._sources:
            yield from source
        yield self

    def _compute_dataset_id(self) -> str:
        """
        Compute the identifier for this `Dataset`. The ID is uniquely defined by the operation(s)
        applied to the `Dataset's` sources.
        """
        return hash_for_serialized_dict({
            "source_ids": [source.id for source in self._sources],
            "logical_op_id": self._operator.get_logical_op_id(),
        })

    def _set_root_datasets(self, new_root_datasets: dict[str, Dataset]) -> None:
        """
        Update the root dataset(s) for this dataset with the `new_root_datasets`. This is used during
        optimization to reuse the same physical plan while running it on a train dataset.

        Args:
            new_root_datasets (dict[str, Dataset]): the new root datasets for this dataset.
        """
        new_sources = []
        for old_source in self._sources:
            if old_source.id in new_root_datasets:
                new_sources.append(new_root_datasets[old_source.id])
            else:
                old_source._set_root_datasets(new_root_datasets)
                new_sources.append(old_source)
        self._sources = new_sources

    # TODO: the entire way (unique) logical op ids are computed and stored needs to be rethought
    def _generate_unique_logical_op_ids(self, topo_idx: int | None = None) -> None:
        """
        Generate unique operation IDs for all operators in this dataset and its sources.
        This is used to ensure that each operator can be uniquely identified during execution.
        """
        # generate the unique op ids for all sources' operators
        for source in self._sources:
            topo_idx = source._generate_unique_logical_op_ids(topo_idx)
            topo_idx += 1

        # if topo_idx is None, this is the first call, so we initialize it to 0
        if topo_idx is None:
            topo_idx = 0

        # compute this operator's unique operator id
        this_unique_logical_op_id = f"{topo_idx}-{self._operator.get_logical_op_id()}"

        # update the unique logical op id for this operator
        self._operator.set_unique_logical_op_id(this_unique_logical_op_id)

        # return the current unique full_op_id for this operator
        return topo_idx

    # TODO
    def _resolve_depends_on(self, depends_on: list[str]) -> list[str]:
        """
        TODO: resolve the `depends_on` strings to their full field names ({Dataset.id}.{field_name}).
        """
        return []

    def _get_root_datasets(self) -> dict[str, Dataset]:
        """Return a mapping from the id --> Dataset for all root datasets."""
        if self.is_root:
            return {self.id: self}

        root_datasets = {}
        for source in self._sources:
            child_root_datasets = source._get_root_datasets()
            root_datasets = {**root_datasets, **child_root_datasets}

        return root_datasets

    def relax_types(self) -> None:
        """
        Relax the types in this Dataset's schema and all upstream Datasets' schemas to be more permissive.
        """
        # relax the types in this dataset's schema
        self._schema = relax_schema(self._schema)

        # relax the types in dataset's operator's input and output schemas
        self._operator.input_schema = None if self._operator.input_schema is None else relax_schema(self._operator.input_schema)
        self._operator.output_schema = relax_schema(self._operator.output_schema)

        # recursively relax the types in all upstream datasets
        for source in self._sources:
            source.relax_types()

    def get_upstream_datasets(self) -> list[Dataset]:
        """
        Get the list of all upstream datasets that are sources to this dataset.
        """
        # recursively get the upstream datasets
        upstream = []
        for source in self._sources:
            upstream.extend(source.get_upstream_datasets())
            upstream.append(source)
        return upstream

    def get_limit(self) -> int | None:
        """Get the limit applied to this Dataset, if any."""
        if isinstance(self._operator, LimitScan):
            return self._operator.limit

        source_limits = []
        for source in self._sources:
            source_limit = source.get_limit()
            if source_limit is not None:
                source_limits.append(source_limit)

        if len(source_limits) == 0:
            return None

        return min([limit for limit in source_limits if limit is not None])

    def copy(self):
        return Dataset(
            sources=[source.copy() for source in self._sources],
            operator=self._operator.copy(),
            schema=self._schema,
            id=self.id,
        )

    def join(self, other: Dataset, on: str | list[str], how: str = "inner") -> Dataset:
        """
        Perform the specified join on the specified (list of) column(s)
        """
        # enforce type for on
        if isinstance(on, str):
            on = [on]

        # construct new output schema
        combined_schema = union_schemas([self.schema, other.schema], join=True, on=on)

        # construct logical operator
        operator = JoinOp(
            input_schema=combined_schema,
            output_schema=combined_schema,
            condition="",
            on=on,
            how=how,
            depends_on=on,
        )

        return Dataset(sources=[self, other], operator=operator, schema=combined_schema)

    def sem_join(self, other: Dataset, condition: str, desc: str | None = None, depends_on: str | list[str] | None = None, how: str = "inner") -> Dataset:
        """
        Perform a semantic (inner) join on the specified join predicate
        """
        # enforce type for depends_on
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        # construct new output schema
        combined_schema = union_schemas([self.schema, other.schema], join=True)

        # construct logical operator
        operator = JoinOp(
            input_schema=combined_schema,
            output_schema=combined_schema,
            condition=condition,
            how=how,
            desc=desc,
            depends_on=depends_on,
        )

        return Dataset(sources=[self, other], operator=operator, schema=combined_schema)

    def filter(
        self,
        filter: Callable,
        depends_on: str | list[str] | None = None,
    ) -> Dataset:
        """Add a user defined function as a filter to the Set. This filter will possibly restrict the items that are returned later."""
        # construct Filter object
        f = None
        if callable(filter):
            f = Filter(filter_fn=filter)
        else:
            error_str = f"Only support callable for filter, currently got {type(filter)}"
            if isinstance(filter, str):
                error_str += ". Consider using sem_filter() for semantic filters."
            raise Exception(error_str)

        # enforce type for depends_on
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        # construct logical operator
        operator = FilteredScan(input_schema=self.schema, output_schema=self.schema, filter=f, depends_on=depends_on)

        return Dataset(sources=[self], operator=operator, schema=self.schema)

    def sem_filter(
        self,
        filter: str,
        desc: str | None = None,
        depends_on: str | list[str] | None = None,
    ) -> Dataset:
        """Add a natural language description of a filter to the Set. This filter will possibly restrict the items that are returned later."""
        # construct Filter object
        f = None
        if isinstance(filter, str):
            f = Filter(filter)
        else:
            raise Exception("sem_filter() only supports `str` input for _filter.", type(filter))

        # enforce type for depends_on
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        # construct logical operator
        operator = FilteredScan(input_schema=self.schema, output_schema=self.schema, filter=f, desc=desc, depends_on=depends_on)

        return Dataset(sources=[self], operator=operator, schema=self.schema)

    def _sem_map(self, cols: list[dict] | type[BaseModel] | None,
                 cardinality: Cardinality,
                 desc: str | None = None,
                 depends_on: str | list[str] | None = None) -> Dataset:
        """Execute the semantic map operation with the appropriate cardinality."""
        # construct new output schema
        new_output_schema = None
        if cols is None:
            new_output_schema = self.schema
        elif isinstance(cols, list):
            cols = create_schema_from_fields(cols)
            new_output_schema = union_schemas([self.schema, cols])
        elif issubclass(cols, BaseModel):
            new_output_schema = union_schemas([self.schema, cols])
        else:
            raise ValueError("`cols` must be a list of dictionaries or a BaseModel.")

        # enforce type for depends_on
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        # construct logical operator
        operator = ConvertScan(
            input_schema=self.schema,
            output_schema=new_output_schema,
            cardinality=cardinality,
            udf=None,
            desc=desc,
            depends_on=depends_on,
        )

        return Dataset(sources=[self], operator=operator, schema=new_output_schema)

    def sem_add_columns(self, cols: list[dict] | type[BaseModel],
                        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
                        desc: str | None = None,
                        depends_on: str | list[str] | None = None) -> Dataset:
        """
        NOTE: we are renaming this function to `sem_map` and deprecating `sem_add_columns` in the next
        release of PZ. To update your code, simply change your calls from `.sem_add_columns(...)` to `.sem_map(...)`.
        The function arguments will stay the same.

        Add new columns by specifying the column names, descriptions, and types.
        The column will be computed during the execution of the Dataset.
        Example:
            sem_add_columns(
                [{'name': 'greeting', 'desc': 'The greeting message', 'type': str},
                 {'name': 'age', 'desc': 'The age of the person', 'type': int},
                 {'name': 'full_name', 'desc': 'The name of the person', 'type': str}]
            )
        """
        # issue deprecation warning
        warnings.warn(
            "we are renaming this function to `sem_map` and deprecating `sem_add_columns` in the next"
            " release of PZ. To update your code, simply change your calls from `.sem_add_columns(...)`"
            " to `.sem_map(...)`. The function arguments will stay the same.",
            DeprecationWarning,
            stacklevel=2
        )

        return self._sem_map(cols, cardinality, desc, depends_on)

    def sem_map(self, cols: list[dict] | type[BaseModel], desc: str | None = None, depends_on: str | list[str] | None = None) -> Dataset:
        """
        Compute new field(s) by specifying their names, descriptions, and types. For each input there will
        be one output. The field(s) will be computed during the execution of the Dataset.

        Example:
            sem_map(
                [{'name': 'greeting', 'desc': 'The greeting message', 'type': str},
                 {'name': 'age', 'desc': 'The age of the person', 'type': int},
                 {'name': 'full_name', 'desc': 'The name of the person', 'type': str}]
            )
        """
        return self._sem_map(cols, Cardinality.ONE_TO_ONE, desc, depends_on)

    def sem_flat_map(self, cols: list[dict] | type[BaseModel], desc: str | None = None, depends_on: str | list[str] | None = None) -> Dataset:
        """
        Compute new field(s) by specifying their names, descriptions, and types. For each input there will
        be one or more output(s). The field(s) will be computed during the execution of the Dataset.

        Example:
            sem_flat_map(
                cols=[
                    {'name': 'author_name', 'description': 'The name of the author', 'type': str},
                    {'name': 'institution', 'description': 'The institution of the author', 'type': str},
                    {'name': 'email', 'description': 'The author's email', 'type': str},
                ]
            )
        """
        return self._sem_map(cols, Cardinality.ONE_TO_MANY, desc, depends_on)

    def _map(self, udf: Callable,
            cols: list[dict] | type[BaseModel] | None,
            cardinality: Cardinality,
            depends_on: str | list[str] | None = None) -> Dataset:
        """Execute the map operation with the appropriate cardinality."""
        # construct new output schema
        new_output_schema = None
        if cols is None:
            new_output_schema = self.schema
        elif isinstance(cols, list):
            cols = create_schema_from_fields(cols)
            new_output_schema = union_schemas([self.schema, cols])
        elif issubclass(cols, BaseModel):
            new_output_schema = union_schemas([self.schema, cols])
        else:
            raise ValueError("`cols` must be a list of dictionaries, a BaseModel, or None.")

        # enforce type for depends_on
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        # construct logical operator
        operator = ConvertScan(
            input_schema=self.schema,
            output_schema=new_output_schema,
            cardinality=cardinality,
            udf=udf,
            depends_on=depends_on,
        )

        return Dataset(sources=[self], operator=operator, schema=new_output_schema)

    def add_columns(self, udf: Callable,
                    cols: list[dict] | type[BaseModel] | None,
                    cardinality: Cardinality = Cardinality.ONE_TO_ONE,
                    depends_on: str | list[str] | None = None) -> Dataset:
        """
        NOTE: we are renaming this function to `map` and deprecating `add_columns` in the next
        release of PZ. To update your code, simply change your calls from `.add_columns(...)` to `.map(...)`.
        The function arguments will stay the same.

        Compute new fields (or update existing ones) with a UDF. For each input, this function will compute one output.

        Set `cols=None` if your add_columns operation is not computing any new fields.

        Examples:
            add_columns(
                udf=compute_personal_greeting,
                cols=[
                    {'name': 'greeting', 'description': 'The greeting message', 'type': str},
                    {'name': 'age', 'description': 'The age of the person', 'type': int},
                    {'name': 'full_name', 'description': 'The name of the person', 'type': str},
                ]
            )
        """
        # issue deprecation warning
        warnings.warn(
            "we are renaming this function to `map` and deprecating `add_columns` in the next"
            " release of PZ. To update your code, simply change your calls from `.add_columns(...)`"
            " to `.map(...)`. The function arguments will stay the same.",
            DeprecationWarning,
            stacklevel=2
        )

        # sanity check inputs
        if udf is None:
            raise ValueError("`udf` must be provided for add_columns.")

        return self._map(udf, cols, cardinality, depends_on)

    def map(self, udf: Callable,
            cols: list[dict] | type[BaseModel] | None,
            depends_on: str | list[str] | None = None) -> Dataset:
        """
        Compute new fields (or update existing ones) with a UDF. For each input, this function will compute one output.

        Set `cols=None` if your map is not computing any new fields.

        Examples:
            map(
                udf=compute_personal_greeting,
                cols=[
                    {'name': 'greeting', 'description': 'The greeting message', 'type': str},
                    {'name': 'age', 'description': 'The age of the person', 'type': int},
                    {'name': 'full_name', 'description': 'The name of the person', 'type': str},
                ]
            )
        """
        # sanity check inputs
        if udf is None:
            raise ValueError("`udf` must be provided for map.")

        return self._map(udf, cols, Cardinality.ONE_TO_ONE, depends_on)

    def flat_map(self, udf: Callable,
            cols: list[dict] | type[BaseModel] | None,
            depends_on: str | list[str] | None = None) -> Dataset:
        """
        Compute new fields (or update existing ones) with a UDF. For each input, this function will compute one or more outputs.

        Set `cols=None` if your flat_map is not computing any new fields.

        Examples:
            flat_map(
                udf=extract_paper_authors,
                cols=[
                    {'name': 'author_name', 'description': 'The name of the author', 'type': str},
                    {'name': 'institution', 'description': 'The institution of the author', 'type': str},
                    {'name': 'email', 'description': 'The author's email', 'type': str},
                ]
            )
        """
        # sanity check inputs
        if udf is None:
            raise ValueError("`udf` must be provided for map.")

        return self._map(udf, cols, Cardinality.ONE_TO_MANY, depends_on)

    def count(self) -> Dataset:
        """Apply a count aggregation to this set"""
        operator = Aggregate(input_schema=self.schema, agg_func=AggFunc.COUNT)
        return Dataset(sources=[self], operator=operator, schema=operator.output_schema)

    def average(self) -> Dataset:
        """Apply an average aggregation to this set"""
        operator = Aggregate(input_schema=self.schema, agg_func=AggFunc.AVERAGE)
        return Dataset(sources=[self], operator=operator, schema=operator.output_schema)

    def sum(self) -> Dataset:
        """Apply a summation to this set"""
        operator = Aggregate(input_schema=self.schema, agg_func=AggFunc.SUM)
        return Dataset(sources=[self], operator=operator, schema=operator.output_schema)

    def min(self) -> Dataset:
        """Apply an min operator to this set"""
        operator = Aggregate(input_schema=self.schema, agg_func=AggFunc.MIN)
        return Dataset(sources=[self], operator=operator, schema=operator.output_schema)

    def max(self) -> Dataset:
        """Apply an max operator to this set"""
        operator = Aggregate(input_schema=self.schema, agg_func=AggFunc.MAX)
        return Dataset(sources=[self], operator=operator, schema=operator.output_schema)

    def groupby(self, groupby: GroupBySig) -> Dataset:
        output_schema = groupby.output_schema()
        operator = GroupByAggregate(input_schema=self.schema, output_schema=output_schema, group_by_sig=groupby)
        return Dataset(sources=[self], operator=operator, schema=output_schema)

    def sem_agg(self, col: dict | type[BaseModel], agg: str, depends_on: str | list[str] | None = None) -> Dataset:
        """
        Apply a semantic aggregation to this set. The `agg` string will be applied using an LLM
        over the entire set of inputs' fields specified in `depends_on` to generate the output `col`.

        Example:
            sem_agg(
                col={'name': 'overall_sentiment', 'desc': 'The overall sentiment of the reviews', 'type': str},
                agg="Compute the overall sentiment of the reviews as POSITIVE or NEGATIVE.",
                depends_on="review_text",
            )
        """
        # construct new output schema
        new_output_schema = None
        if isinstance(col, dict):
            col_schema = create_schema_from_fields([col])
            new_output_schema = union_schemas([self.schema, col_schema])
        elif issubclass(col, BaseModel):
            assert len(col.model_fields) == 1, "For semantic aggregation, when passing a BaseModel to `col` it must have exactly one field."
            new_output_schema = union_schemas([self.schema, col])
        else:
            raise ValueError("`col` must be a dictionary or a single-field BaseModel.")

        # enforce type for depends_on
        if isinstance(depends_on, str):
            depends_on = [depends_on]

        # construct logical operator
        operator = Aggregate(input_schema=self.schema, output_schema=new_output_schema, agg_str=agg, depends_on=depends_on)

        return Dataset(sources=[self], operator=operator, schema=operator.output_schema)

    def sem_topk(
        self,
        index: Collection,
        search_attr: str,
        output_attrs: list[dict] | type[BaseModel],
        search_func: Callable | None = None,
        k: int = -1,
    ) -> Dataset:
        """
        Retrieve the top-k nearest neighbors of the value of the `search_attr` from the `index` and
        use these results to construct the `output_attrs` field(s).
        """
        # construct new output schema
        new_output_schema = None
        if isinstance(output_attrs, list):
            output_attrs = create_schema_from_fields(output_attrs)
            new_output_schema = union_schemas([self.schema, output_attrs])
        elif issubclass(output_attrs, BaseModel):
            new_output_schema = union_schemas([self.schema, output_attrs])
        else:
            raise ValueError("`output_attrs` must be a list of dictionaries or a BaseModel.")

        # TODO: revisit once we can think through abstraction(s)
        # # construct the PZIndex from the user-provided index
        # index = index_factory(index)

        # construct logical operator
        operator = TopKScan(
            input_schema=self.schema,
            output_schema=new_output_schema,
            index=index,
            search_func=search_func,
            search_attr=search_attr,
            output_attrs=output_attrs,
            k=k,
        )

        return Dataset(sources=[self], operator=operator, schema=new_output_schema)

    def limit(self, n: int) -> Dataset:
        """Limit the set size to no more than n rows"""
        operator = LimitScan(input_schema=self.schema, output_schema=self.schema, limit=n)
        return Dataset(sources=[self], operator=operator, schema=self.schema)

    def distinct(self, distinct_cols: list[str] | None = None) -> Dataset:
        """Return a new Dataset with distinct rows based on the current schema."""
        operator = Distinct(input_schema=self.schema, output_schema=self.schema, distinct_cols=distinct_cols)
        return Dataset(sources=[self], operator=operator, schema=self.schema)

    def project(self, project_cols: list[str] | str) -> Dataset:
        """Project the Set to only include the specified columns."""
        project_cols = project_cols if isinstance(project_cols, list) else [project_cols]
        new_output_schema = project(self.schema, project_cols)
        operator = Project(input_schema=self.schema, output_schema=new_output_schema, project_cols=project_cols)
        return Dataset(sources=[self], operator=operator, schema=new_output_schema)

    def run(self, config: QueryProcessorConfig | None = None, **kwargs):
        """Invoke the QueryProcessor to execute the query. `kwargs` will be applied to the QueryProcessorConfig."""
        # TODO: this import currently needs to be here to avoid a circular import; we should fix this in a subsequent PR
        from palimpzest.query.processor.query_processor_factory import QueryProcessorFactory

        # as syntactic sugar, we will allow some keyword arguments to parameterize our policies
        policy = construct_policy_from_kwargs(**kwargs)
        if policy is not None:
            kwargs["policy"] = policy

        # construct unique logical op ids for all operators in this dataset
        self._generate_unique_logical_op_ids()

        return QueryProcessorFactory.create_and_run_processor(self, config)

    def optimize_and_run(self, config: QueryProcessorConfig | None = None, train_dataset: dict[str, Dataset] | Dataset | None = None, validator: Validator | None = None, **kwargs):
        """Optimize the PZ program using the train_dataset and validator before running the optimized plan."""
        # TODO: this import currently needs to be here to avoid a circular import; we should fix this in a subsequent PR
        from palimpzest.query.processor.query_processor_factory import QueryProcessorFactory

        # confirm that either train_dataset or validator is provided
        assert train_dataset is not None or validator is not None, "Must provide at least one of train_dataset or validator to use optimize_and_run()"

        # validate the train_dataset has one input for each source dataset and normalize its type to be a dict
        if train_dataset is not None:
            root_datasets = self._get_root_datasets()
            if isinstance(train_dataset, Dataset) and len(root_datasets) > 1:
                raise ValueError(
                    "For plans with more than one root dataset, `train_dataset` must be a dictionary mapping"
                    " {'dataset_id' --> Dataset} for all root Datasets"
                )

            elif isinstance(train_dataset, Dataset):
                root_dataset_id = list(root_datasets.values())[0].id
                if train_dataset.id != root_dataset_id:
                    warnings.warn(
                        f"train_dataset.id={train_dataset.id} does not match root dataset id={root_dataset_id}\n"
                        f"Setting train_dataset to be the training data for root dataset with id={root_dataset_id} anyways.",
                        stacklevel=2,
                    )
                train_dataset = {root_dataset_id: train_dataset}

            elif not all(dataset_id in train_dataset for dataset_id in root_datasets):
                missing_ids = [dataset_id for dataset_id in root_datasets if dataset_id not in train_dataset]
                raise ValueError(
                    f"`train_dataset` is missing the following root dataset id(s): {missing_ids}"
                )

        # as syntactic sugar, we will allow some keyword arguments to parameterize our policies
        policy = construct_policy_from_kwargs(**kwargs)
        if policy is not None:
            kwargs["policy"] = policy

        # construct unique logical op ids for all operators in this dataset
        self._generate_unique_logical_op_ids()

        return QueryProcessorFactory.create_and_run_processor(self, config, train_dataset, validator)



import logging

from palimpzest.core.elements.records import DataRecord
from palimpzest.core.models import PlanStats
from palimpzest.query.execution.execution_strategy import ExecutionStrategy
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.join import JoinOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.scan import ContextScanOp, ScanPhysicalOp
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.utils.progress import create_progress_manager

logger = logging.getLogger(__name__)

class SequentialSingleThreadExecutionStrategy(ExecutionStrategy):
    """
    A single-threaded execution strategy that processes operators sequentially.
    
    This strategy processes all records through one operator completely before moving to the next operator
    in the execution plan. For example, if we have operators A -> B -> C and records [1,2,3]:
    1. First processes records [1,2,3] through operator A
    2. Then takes A's output and processes all of it through operator B
    3. Finally processes all of B's output through operator C
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = 1

    def _execute_plan(self, plan: PhysicalPlan, input_queues: dict[str, dict[str, list]], plan_stats: PlanStats) -> tuple[list[DataRecord], PlanStats]:
        # execute the plan one operator at a time
        output_records = []
        for topo_idx, operator in enumerate(plan):
            # if we've filtered out all records, terminate early
            source_unique_full_op_ids = (
                [f"source_{operator.get_full_op_id()}"]
                if isinstance(operator, (ContextScanOp, ScanPhysicalOp))
                else plan.get_source_unique_full_op_ids(topo_idx, operator)
            )
            unique_full_op_id = f"{topo_idx}-{operator.get_full_op_id()}"
            num_inputs = sum(len(input_queues[unique_full_op_id][source_unique_full_op_id]) for source_unique_full_op_id in source_unique_full_op_ids)
            if num_inputs == 0:
                break

            # begin to process this operator
            records, record_op_stats = [], []
            logger.info(f"Processing operator {operator.op_name()} ({unique_full_op_id})")

            # if this operator is an aggregate, process all the records in the input_queue
            if isinstance(operator, AggregateOp):
                source_unique_full_op_id = source_unique_full_op_ids[0]
                record_set = operator(candidates=input_queues[unique_full_op_id][source_unique_full_op_id])
                records = record_set.data_records
                record_op_stats = record_set.record_op_stats
                num_outputs = sum(record._passed_operator for record in records)

                # update the progress manager
                self.progress_manager.incr(unique_full_op_id, num_inputs=1, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

            # if this operator is a join, process all pairs of records from the two input queues
            elif isinstance(operator, JoinOp):
                left_full_source_op_id = source_unique_full_op_ids[0]
                left_num_inputs = len(input_queues[unique_full_op_id][left_full_source_op_id])
                left_input_records = [input_queues[unique_full_op_id][left_full_source_op_id].pop(0) for _ in range(left_num_inputs)]

                right_full_source_op_id = source_unique_full_op_ids[1]
                right_num_inputs = len(input_queues[unique_full_op_id][right_full_source_op_id])
                right_input_records = [input_queues[unique_full_op_id][right_full_source_op_id].pop(0) for _ in range(right_num_inputs)]

                record_set, num_inputs_processed = operator(left_input_records, right_input_records)
                records = record_set.data_records
                record_op_stats = record_set.record_op_stats

                # process the join one last time with final=True to handle any left/right/outer join logic
                if operator.how in ("left", "right", "outer"):
                    record_set, num_inputs_processed = operator([], [], final=True)
                    records.extend(record_set.data_records)
                    record_op_stats.extend(record_set.record_op_stats)
      
                num_outputs = sum(record._passed_operator for record in records)

                # update the progress manager
                self.progress_manager.incr(unique_full_op_id, num_inputs=num_inputs_processed, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

            # otherwise, process the records in the input queue for this operator one at a time
            else:
                source_unique_full_op_id = source_unique_full_op_ids[0]
                for input_record in input_queues[unique_full_op_id][source_unique_full_op_id]:
                    record_set = operator(input_record)
                    records.extend(record_set.data_records)
                    record_op_stats.extend(record_set.record_op_stats)
                    num_outputs = sum(record._passed_operator for record in record_set.data_records)

                    # update the progress manager
                    self.progress_manager.incr(unique_full_op_id, num_inputs=1, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                    # finish early if this is a limit
                    if isinstance(operator, LimitScanOp) and len(records) == operator.limit:
                        break

            # update plan stats
            plan_stats.add_record_op_stats(unique_full_op_id, record_op_stats)

            # update next input_queue (if it exists)
            output_records = [record for record in records if record._passed_operator]
            next_unique_full_op_id = plan.get_next_unique_full_op_id(topo_idx, operator)
            if next_unique_full_op_id is not None:
                input_queues[next_unique_full_op_id][unique_full_op_id] = output_records

            logger.info(f"Finished processing operator {operator.op_name()} ({unique_full_op_id}), and generated {len(records)} records")

        # finalize plan stats
        plan_stats.finish()

        return output_records, plan_stats

    def execute_plan(self, plan: PhysicalPlan) -> tuple[list[DataRecord], PlanStats]:
        """Initialize the stats and execute the plan."""
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues for each operation
        input_queues = self._create_input_queues(plan)

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, num_samples=self.num_samples, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _execute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats


class PipelinedSingleThreadExecutionStrategy(ExecutionStrategy):
    """
    A single-threaded execution strategy that processes records through a pipeline of operators.
    
    This strategy implements a pipelined execution model where each record flows through
    the entire operator chain before the next record is processed.

    Example Flow:
    For operators A -> B -> C and records [1,2,3]:
    1. Record 1: A -> B -> C
    2. Record 2: A -> B -> C
    3. Record 3: A -> B -> C
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = 1

    def _any_queue_not_empty(self, queues: dict[str, list] | dict[str, dict[str, list]]) -> bool:
        """Helper function to check if any queue is not empty."""
        for _, value in queues.items():
            if isinstance(value, dict):
                if any(len(subqueue) > 0 for subqueue in value.values()):
                    return True
            elif len(value) > 0:
                return True
        return False

    def _upstream_ops_finished(self, plan: PhysicalPlan, unique_full_op_id: str, input_queues: dict[str, dict[str, list]]) -> bool:
        """Helper function to check if agg / join operator is ready to process its inputs."""
        upstream_unique_full_op_ids = plan.get_upstream_unique_full_op_ids(unique_full_op_id)
        upstream_input_queues = {upstream_unique_full_op_id: input_queues[upstream_unique_full_op_id] for upstream_unique_full_op_id in upstream_unique_full_op_ids}
        return not self._any_queue_not_empty(upstream_input_queues)


    def _execute_plan(self, plan: PhysicalPlan, input_queues: dict[str, dict[str, list]], plan_stats: PlanStats) -> tuple[list[DataRecord], PlanStats]:
        # execute the plan until either:
        # 1. all records have been processed, or
        # 2. the final limit operation has completed (we break out of the loop if this happens)
        final_output_records = []
        while self._any_queue_not_empty(input_queues):
            for topo_idx, operator in enumerate(plan):
                # if this operator does not have enough inputs to execute, then skip it
                source_unique_full_op_ids = (
                    [f"source_{operator.get_full_op_id()}"]
                    if isinstance(operator, (ContextScanOp, ScanPhysicalOp))
                    else plan.get_source_unique_full_op_ids(topo_idx, operator)
                )
                unique_full_op_id = f"{topo_idx}-{operator.get_full_op_id()}"

                num_inputs = sum(len(input_queues[unique_full_op_id][source_unique_full_op_id]) for source_unique_full_op_id in source_unique_full_op_ids)
                agg_op_not_ready = isinstance(operator, AggregateOp) and not self._upstream_ops_finished(plan, unique_full_op_id, input_queues)
                join_op_not_ready = isinstance(operator, JoinOp) and not self._upstream_ops_finished(plan, unique_full_op_id, input_queues)
                if num_inputs == 0 or agg_op_not_ready or join_op_not_ready:
                    continue

                # create empty lists for records and execution stats generated by executing this operator on its next input(s)
                records, record_op_stats = [], []

                # if the next operator is an aggregate, process all the records in the input_queue
                if isinstance(operator, AggregateOp):
                    source_unique_full_op_id = source_unique_full_op_ids[0]
                    input_records = [input_queues[unique_full_op_id][source_unique_full_op_id].pop(0) for _ in range(num_inputs)]
                    record_set = operator(candidates=input_records)
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats
                    num_outputs = sum(record._passed_operator for record in records)

                    # update the progress manager
                    self.progress_manager.incr(unique_full_op_id, num_inputs=1, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                # if this operator is a join, process all pairs of records from the two input queues
                elif isinstance(operator, JoinOp):
                    left_full_source_op_id = source_unique_full_op_ids[0]
                    left_num_inputs = len(input_queues[unique_full_op_id][left_full_source_op_id])
                    left_input_records = [input_queues[unique_full_op_id][left_full_source_op_id].pop(0) for _ in range(left_num_inputs)]

                    right_full_source_op_id = source_unique_full_op_ids[1]
                    right_num_inputs = len(input_queues[unique_full_op_id][right_full_source_op_id])
                    right_input_records = [input_queues[unique_full_op_id][right_full_source_op_id].pop(0) for _ in range(right_num_inputs)]

                    record_set, num_inputs_processed = operator(left_input_records, right_input_records)
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats
                    num_outputs = sum(record._passed_operator for record in records)

                    # update the progress manager
                    self.progress_manager.incr(unique_full_op_id, num_inputs=num_inputs_processed, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                # otherwise, process the next record in the input queue for this operator
                else:
                    source_unique_full_op_id = source_unique_full_op_ids[0]
                    input_record = input_queues[unique_full_op_id][source_unique_full_op_id].pop(0)
                    record_set = operator(input_record)
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats
                    num_outputs = sum(record._passed_operator for record in records)

                    # update the progress manager
                    self.progress_manager.incr(unique_full_op_id, num_inputs=1, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                # if this is a join operator with no more inputs to process, then finish it
                if isinstance(operator, JoinOp) and operator.how in ("left", "right", "outer"):
                    join_op_upstream_finished = self._upstream_ops_finished(plan, unique_full_op_id, input_queues)
                    join_input_queues_empty = all(len(inputs) == 0 for inputs in input_queues[unique_full_op_id].values())
                    if join_op_upstream_finished and join_input_queues_empty and not operator.finished:
                        # process the join one last time with final=True to handle any left/right/outer join logic
                        record_set, num_inputs_processed = operator([], [], final=True)
                        records.extend(record_set.data_records)
                        record_op_stats.extend(record_set.record_op_stats)
                        num_outputs += sum(record._passed_operator for record in record_set.data_records)
                        operator.set_finished()

                # update plan stats
                plan_stats.add_record_op_stats(unique_full_op_id, record_op_stats)

                # update next input_queue or final_output_records
                output_records = [record for record in records if record._passed_operator]
                next_unique_full_op_id = plan.get_next_unique_full_op_id(topo_idx, operator)
                if next_unique_full_op_id is not None:
                    input_queues[next_unique_full_op_id][unique_full_op_id].extend(output_records)
                else:
                    final_output_records.extend(output_records)

                logger.info(f"Finished processing operator {operator.op_name()} ({unique_full_op_id}) on {num_inputs} records")

            # break out of loop if the final operator is a LimitScanOp and we've reached its limit
            if isinstance(plan.operator, LimitScanOp) and len(final_output_records) == plan.operator.limit:
                break

        # finalize plan stats
        plan_stats.finish()

        return final_output_records, plan_stats

    def execute_plan(self, plan: PhysicalPlan):
        """Initialize the stats and execute the plan."""
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues for each operation
        input_queues = self._create_input_queues(plan)        

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, self.num_samples, self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _execute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats



### MAP ###
MAP_VALIDATOR_PROMPT = """You are an intelligent judge whose job is to evaluate how successfully an agent executed a given instruction.
You will be presented with the input(s) provided to the agent followed by the output produced by the agent.

Each output will be a dictionary. The keys will be **output fields** which were computed by the agent.

Your job will be to assign a score of 1.0 to every output field which was computed correctly, and a score of 0.0 to every output field which was computed incorrectly. If the output for a field is a list, you may give a score in between 0.0 and 1.0 representing the fraction of correct items in the list.

Here is an example evaluation:

INPUT MESSAGES:
---------------
You are a helpful assistant whose job is to generate a JSON object. You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

INPUT FIELDS:
- text: a text passage describing a scientist
- birthday: the scientist's birthday

OUTPUT FIELDS:
- name: the name of the scientist
- birth_year: the year the scientist was born

CONTEXT:
{
  "text": "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace, was an English mathematician and writer chiefly known for her work on Charles Babbage's proposed mechanical general-purpose computer, the Analytical Engine. She was the first to recognise that the machine had applications beyond pure calculation.",
  "birthday": "December 10, 1815"
}

OUTPUT:
--------
{
  "name": "Charles Babbage",
  "birth_year": 1815
}

EVALUATION: {"name": 0.0, "birth_year": 1.0}

Remember, be sure to output your evaluation as a dictionary where each value contains a 0.0 or 1.0 score for each output field (or a score within [0.0, 1.0] for list output fields).

INPUT MESSAGES:
---------------

"""

MAP_IMAGE_VALIDATOR_PROMPT = """You are an intelligent judge whose job is to evaluate how successfully an agent executed a given instruction.
You will be presented with the input(s) provided to the agent followed by the output produced by the agent.

Each output will be a dictionary. The keys will be **output fields** which were computed by the agent.

Your job will be to assign a score of 1.0 to every output field which was computed correctly, and a score of 0.0 to every output field which was computed incorrectly. If the output for a field is a list, you may give a score in between 0.0 and 1.0 representing the fraction of correct items in the list.

Here is an example evaluation:

INPUT MESSAGES:
---------------
You are a helpful assistant whose job is to analyze input image(s) and/or text in order to produce a JSON object. You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

INPUT FIELDS:
- image: an image of a scene
- photographer: the photographer of the image

OUTPUT FIELDS:
- dog_in_image: true if a dog is in the image and false otherwise
- person_in_image: true if a person is in the image and false otherwise

CONTEXT:
{
  "image": <bytes>,
  "photographer": "CameraEnthusiast1"
}
<image content provided here; assume in this example the image shows a dog and a cat playing>

OUTPUT:
--------
{
  "dog_in_image": true,
  "person_in_image": true
}

EVALUATION: {"dog_in_image": 1.0, "person_in_image": 0.0}

Remember, be sure to output your evaluation as a dictionary where each value contains a 0.0 or 1.0 score for each output field (or a score within [0.0, 1.0] for list output fields).

INPUT MESSAGES:
---------------

"""


### FLAT MAP ###
FLAT_MAP_VALIDATOR_PROMPT = """You are an intelligent judge whose job is to evaluate how successfully an agent executed a given instruction.
You will be presented with the input(s) provided to the agent followed by the output(s) produced by the agent.

Each output will be a list of dictionaries. The keys of each dictionary will be **output fields** which were computed by the agent.

Your job will be to assign a score of 1.0 to every output field which was computed correctly, and a score of 0.0 to every output field which was computed incorrectly. If the output for a field is a list, you may give a score in between 0.0 and 1.0 representing the fraction of correct items in the list.

Here is an example evaluation:

INPUT MESSAGES:
---------------
You are a helpful assistant whose job is to generate a JSON object. You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

INPUT FIELDS:
- text: a text passage describing scientists
- birthdays: text containing birth dates

OUTPUT FIELDS:
- name: the name of the scientist
- birth_year: the year the scientist was born

CONTEXT:
{
  "text": "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace, was an English mathematician and writer chiefly known for her work on Charles Babbage's proposed mechanical general-purpose computer, the Analytical Engine. She was the first to recognise that the machine had applications beyond pure calculation.",
  "birthdays": "...Lovelace was born on December 10, 1815, almost exactly 24 years after Babbage's birth on 26 December 1791..."
}

OUTPUTS:
--------
[
  {
    "name": "Ada Lovelace",
    "birth_year": 1815
  },
  {
    "name": "Charles Babbage",
    "birth_year": 1790
  }
]

EVALUATION: [{"name": 1.0, "birth_year": 1.0}, {"name": 1.0, "birth_year": 0.0}]

Remember, be sure to output your evaluation as a list of dictionaries where each dictionary contains a 0.0 or 1.0 score for each output field (or a score within [0.0, 1.0] for list output fields).

INPUT MESSAGES:
---------------

"""

FLAT_MAP_IMAGE_VALIDATOR_PROMPT = """You are an intelligent judge whose job is to evaluate how successfully an agent executed a given instruction.
You will be presented with the input(s) provided to the agent followed by the output(s) produced by the agent.

Each output will be a list of dictionaries. The keys of each dictionary will be **output fields** which were computed by the agent.

Your job will be to assign a score of 1.0 to every output field which was computed correctly, and a score of 0.0 to every output field which was computed incorrectly. If the output for a field is a list, you may give a score in between 0.0 and 1.0 representing the fraction of correct items in the list.

Here is an example evaluation:

INPUT MESSAGES:
---------------
You are a helpful assistant whose job is to analyze input image(s) and/or text in order to produce a JSON object. You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

INPUT FIELDS:
- image: an image of a scene
- photographer: the photographer of the image

OUTPUT FIELDS:
- animal: the type of animal in the image
- animal_is_canine: true if the animal is a canine and false otherwise

CONTEXT:
{
  "image": <bytes>,
  "photographer": "CameraEnthusiast1"
}
<image content provided here; assume in this example the image shows a dog and a cat playing>

OUTPUT:
--------
[
  {
    "animal": "dog",
    "animal_is_canine": true
  },
  {
    "animal": "cat",
    "animal_is_canine": true
  }
]

EVALUATION: [{"animal": 1.0, "animal_is_canine": 1.0}, {"animal": 1.0, "animal_is_canine": 0.0}]

Remember, be sure to output your evaluation as a list of dictionaries where each dictionary contains a 0.0 or 1.0 score for each output field (or a score within [0.0, 1.0] for list output fields).

INPUT MESSAGES:
---------------

"""


### RETRIEVE
RETRIEVE_VALIDATOR_PROMPT = """You are an intelligent judge whose job is to evaluate how successfully an agent executed a given instruction.
You will be presented with the input(s) provided to the agent followed by the output produced by the agent.

Each output will be a dictionary. The keys will be **output fields** which were computed by the agent.

Your job will be to assign a score of 1.0 to every output field which was computed correctly, and a score of 0.0 to every output field which was computed incorrectly. If the output for a field is a list, you may give a score in between 0.0 and 1.0 representing the fraction of correct items in the list.

Here is an example evaluation:

INPUT MESSAGES:
---------------
You are a helpful assistant whose job is to generate a JSON object. You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

INPUT FIELDS:
- text: a text passage describing a scientist

OUTPUT FIELDS:
- related_scientists: list of scientists who perform similar work as the scientist described in the text

CONTEXT:
{
  "text": "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace, was an English mathematician and writer chiefly known for her work on Charles Babbage's proposed mechanical general-purpose computer, the Analytical Engine. She was the first to recognise that the machine had applications beyond pure calculation.",
}

OUTPUT:
--------
{
  "related_scientists": [
    "Charles Babbage",
    "Alan Turing",
    "Charles Darwin",
    "John von Neumann",
  ]
}

EVALUATION: {"related_scientists": 0.75}

Remember, be sure to output your evaluation as a dictionary where each value contains a 0.0 or 1.0 score for each output field (or a score within [0.0, 1.0] for list output fields).

INPUT MESSAGES:
---------------

"""



from palimpzest.query.optimizer.rules import AddContextsBeforeComputeRule as _AddContextsBeforeComputeRule
from palimpzest.query.optimizer.rules import (
    AggregateRule as _AggregateRule,
)
from palimpzest.query.optimizer.rules import (
    BasicSubstitutionRule as _BasicSubstitutionRule,
)
from palimpzest.query.optimizer.rules import (
    CritiqueAndRefineRule as _CritiqueAndRefineRule,
)
from palimpzest.query.optimizer.rules import (
    EmbeddingJoinRule as _EmbeddingJoinRule,
)
from palimpzest.query.optimizer.rules import (
    ImplementationRule as _ImplementationRule,
)
from palimpzest.query.optimizer.rules import (
    LLMConvertBondedRule as _LLMConvertBondedRule,
)
from palimpzest.query.optimizer.rules import (
    LLMFilterRule as _LLMFilterRule,
)
from palimpzest.query.optimizer.rules import (
    MixtureOfAgentsRule as _MixtureOfAgentsRule,
)
from palimpzest.query.optimizer.rules import (
    NestedLoopsJoinRule as _NestedLoopsJoinRule,
)
from palimpzest.query.optimizer.rules import (
    NonLLMConvertRule as _NonLLMConvertRule,
)
from palimpzest.query.optimizer.rules import (
    NonLLMFilterRule as _NonLLMFilterRule,
)
from palimpzest.query.optimizer.rules import (
    PushDownFilter as _PushDownFilter,
)
from palimpzest.query.optimizer.rules import (
    RAGRule as _RAGRule,
)
from palimpzest.query.optimizer.rules import (
    RelationalJoinRule as _RelationalJoinRule,
)
from palimpzest.query.optimizer.rules import (
    ReorderConverts as _ReorderConverts,
)
from palimpzest.query.optimizer.rules import (
    Rule as _Rule,
)
from palimpzest.query.optimizer.rules import (
    SemanticAggregateRule as _SemanticAggregateRule,
)
from palimpzest.query.optimizer.rules import (
    SplitRule as _SplitRule,
)
from palimpzest.query.optimizer.rules import (
    TopKRule as _TopKRule,
)
from palimpzest.query.optimizer.rules import (
    TransformationRule as _TransformationRule,
)

ALL_RULES = [
    _AddContextsBeforeComputeRule,
    _AggregateRule,
    _BasicSubstitutionRule,
    _CritiqueAndRefineRule,
    _EmbeddingJoinRule,
    _ImplementationRule,
    _LLMConvertBondedRule,
    _LLMFilterRule,
    _NestedLoopsJoinRule,
    _MixtureOfAgentsRule,
    _NonLLMConvertRule,
    _NonLLMFilterRule,
    _PushDownFilter,
    _RAGRule,
    _RelationalJoinRule,
    _ReorderConverts,
    _TopKRule,
    _Rule,
    _SemanticAggregateRule,
    _SplitRule,
    _TransformationRule,
]

IMPLEMENTATION_RULES = [
    rule
    for rule in ALL_RULES
    if issubclass(rule, _ImplementationRule)
    and rule not in [_ImplementationRule]
]

TRANSFORMATION_RULES = [
    rule for rule in ALL_RULES if issubclass(rule, _TransformationRule) and rule not in [_TransformationRule]
]

__all__ = [
    "ALL_RULES",
    "IMPLEMENTATION_RULES",
    "TRANSFORMATION_RULES",
]



from __future__ import annotations

import logging
from abc import ABC, abstractmethod

from palimpzest.policy import Policy
from palimpzest.query.optimizer.plan import PhysicalPlan, SentinelPlan
from palimpzest.query.optimizer.primitives import Group

logger = logging.getLogger(__name__)


class OptimizationStrategy(ABC):
    @abstractmethod
    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[PhysicalPlan] | list[SentinelPlan]:
        """Strategy decides how to search through the groups for optimal plan(s)"""
        pass


class GreedyStrategy(OptimizationStrategy):
    def _get_greedy_physical_plan(self, groups: dict, group_id: int) -> PhysicalPlan:
        """
        Return the best plan with respect to the user provided policy.
        """
        # get the best physical expression for this group
        best_phys_expr = groups[group_id].best_physical_expression

        # if this expression has no inputs (i.e. it is a BaseScan), create and return the physical plan
        best_plan = None
        if len(best_phys_expr.input_group_ids) == 0:
            best_plan = PhysicalPlan(best_phys_expr.operator, subplans=None, plan_cost=best_phys_expr.plan_cost)

        # otherwise, if this expression is not a join (i.e. it has one input)
        elif len(best_phys_expr.input_group_ids) == 1:
            # get the best physical plan for this group's input
            input_group_id = best_phys_expr.input_group_ids[0]
            input_best_phys_plan = self._get_greedy_physical_plan(groups, input_group_id)

            # add this operator to best physical plan and return
            best_plan = PhysicalPlan(best_phys_expr.operator, subplans=[input_best_phys_plan], plan_cost=best_phys_expr.plan_cost)

        # otherwise, this expression is a join (i.e. it has two inputs)
        elif len(best_phys_expr.input_group_ids) == 2:
            left_input_group_id, right_input_group_id = best_phys_expr.input_group_ids

            # get the best physical plan for the left input
            left_best_phys_plan = self._get_greedy_physical_plan(groups, left_input_group_id)

            # get the best physical plan for the right input
            right_best_phys_plan = self._get_greedy_physical_plan(groups, right_input_group_id)

            # add this operator to best physical plan and return
            best_plan = PhysicalPlan(best_phys_expr.operator, subplans=[left_best_phys_plan, right_best_phys_plan], plan_cost=best_phys_expr.plan_cost)

        # add this operator to best physical plan and return
        return best_plan

    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[PhysicalPlan]:
        logger.info(f"Getting greedy optimal plans for final group id: {final_group_id}")
        plans = [self._get_greedy_physical_plan(groups, final_group_id)]
        logger.info(f"Done getting greedy optimal plans for final group id: {final_group_id}")

        return plans


class ParetoStrategy(OptimizationStrategy):
    def _get_candidate_pareto_physical_plans(self, groups: dict, group_id: int, policy: Policy) -> list[PhysicalPlan]:
        """
        Return a list of plans which will contain all of the pareto optimal plans (and some additional
        plans which may not be pareto optimal).

        TODO: can we cache group_id --> final_pareto_optimal_plans to avoid re-computing upstream
        groups' pareto-optimal plans for each expression?
        """
        # get the pareto optimal physical expressions for this group
        pareto_optimal_phys_exprs = groups[group_id].pareto_optimal_physical_expressions

        # construct list of pareto optimal plans
        pareto_optimal_plans = []
        for phys_expr in pareto_optimal_phys_exprs:
            # if this expression has no inputs (i.e. it is a BaseScan), create and return the physical plan
            if len(phys_expr.input_group_ids) == 0:
                for plan_cost, _ in phys_expr.pareto_optimal_plan_costs:
                    plan = PhysicalPlan(phys_expr.operator, subplans=None, plan_cost=plan_cost)
                    pareto_optimal_plans.append(plan)

            # otherwise, if this expression is not a join (i.e. it has one input)
            elif len(phys_expr.input_group_ids) == 1:
                # get the pareto optimal physical plan(s) for this group's inputs
                input_group_id = phys_expr.input_group_ids[0]
                pareto_optimal_phys_subplans = self._get_candidate_pareto_physical_plans(groups, input_group_id, policy)

                # iterate over the input subplans and find the one(s) which combine with this physical expression
                # to make a pareto-optimal plan
                for plan_cost, (input_plan_cost, _) in phys_expr.pareto_optimal_plan_costs:
                    for subplan in pareto_optimal_phys_subplans:
                        if subplan.plan_cost == input_plan_cost:
                            plan = PhysicalPlan(phys_expr.operator, subplans=[subplan], plan_cost=plan_cost)
                            pareto_optimal_plans.append(plan)

            # otherwise, this expression is a join (i.e. it has two inputs)
            elif len(phys_expr.input_group_ids) == 2:
                left_input_group_id, right_input_group_id = phys_expr.input_group_ids
                pareto_optimal_left_subplans = self._get_candidate_pareto_physical_plans(groups, left_input_group_id, policy)
                pareto_optimal_right_subplans = self._get_candidate_pareto_physical_plans(groups, right_input_group_id, policy)

                # iterate over the input subplans and find the one(s) which combine with this physical expression
                # to make a pareto-optimal plan
                for plan_cost, (left_input_plan_cost, right_input_plan_cost) in phys_expr.pareto_optimal_plan_costs:
                    for left_subplan in pareto_optimal_left_subplans:
                        if left_subplan.plan_cost == left_input_plan_cost:
                            for right_subplan in pareto_optimal_right_subplans:
                                if right_subplan.plan_cost == right_input_plan_cost:
                                    plan = PhysicalPlan(phys_expr.operator, subplans=[left_subplan, right_subplan], plan_cost=plan_cost)
                                    pareto_optimal_plans.append(plan)

        return pareto_optimal_plans

    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[PhysicalPlan]:
        logger.info(f"Getting pareto optimal plans for final group id: {final_group_id}")
        # compute all of the pareto optimal physical plans
        plans = self._get_candidate_pareto_physical_plans(groups, final_group_id, policy)

        # adjust plans' plan_cost.quality to reflect only the quality of the final operator
        if use_final_op_quality:
            for plan in plans:
                plan.plan_cost.quality = plan.plan_cost.op_estimates.quality

        # filter pareto optimal plans for ones which satisfy policy constraint (if at least one of them does)
        if any([policy.constraint(plan.plan_cost) for plan in plans]):
            plans = [plan for plan in plans if policy.constraint(plan.plan_cost)]

        # select the plan which is best for the given policy
        optimal_plan, plans = plans[0], plans[1:]
        for plan in plans:
            optimal_plan = optimal_plan if policy.choose(optimal_plan.plan_cost, plan.plan_cost) else plan

        plans = [optimal_plan]
        logger.info(f"Done getting pareto optimal plans for final group id: {final_group_id}")
        return plans
    

class SentinelStrategy(OptimizationStrategy):
    def _get_sentinel_plan(self, groups: dict[str, Group], group_id: int) -> SentinelPlan:
        """
        Create and return a SentinelPlan object.

        NOTE: this strategy is only used to construct a SentinelPlan before performing optimization.
              Currently, we do not perform any transformation rules when building the groups which
              are fed into this function. Thus, every physical expression will correspond to the same
              logical operator and share the same logical_op_id. Eventually we will want to consider
              multiple logical re-orderings of operators in our SentinelPlan, but for now it is static.
        """
        # get all the physical expressions for this group as well as their logical_op_id
        phys_exprs = groups[group_id].physical_expressions
        phys_op_set = [expr.operator for expr in phys_exprs]

        # if this expression has no inputs (i.e. it is a scan operator), create and return the sentinel plan
        best_phys_expr = groups[group_id].best_physical_expression
        if len(best_phys_expr.input_group_ids) == 0:
            return SentinelPlan(operator_set=phys_op_set, subplans=None)

        # get the subplans
        subplans = []
        for input_group_id in best_phys_expr.input_group_ids:
            subplan = self._get_sentinel_plan(groups, input_group_id)
            subplans.append(subplan)

        # compose the current physical operator set with its subplans
        return SentinelPlan(operator_set=phys_op_set, subplans=subplans)

    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[SentinelPlan]:
        logger.info(f"Getting sentinel optimal plans for final group id: {final_group_id}")
        plans = [self._get_sentinel_plan(groups, final_group_id)]
        logger.info(f"Done getting sentinel optimal plans for final group id: {final_group_id}")
        return plans


class NoOptimizationStrategy(GreedyStrategy):
    """
    NoOptimizationStrategy is used to intentionally construct a PhysicalPlan without applying any
    logical transformations or optimizations. It uses the same get_optimal_plans logic as the
    GreedyOptimizationStrategy.
    """



"""This file contains prompts for aggregation operations."""

### BASE PROMPTS ###
AGG_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and an output field to generate. Your task is to generate a JSON object which aggregates the input and fills in the output field with the correct value.
You will be provided with a description of each input field and each output field. The field in the output JSON object can be derived using information from the context.

{output_format_instruction} Finish your response with a newline character followed by ---

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

OUTPUT FIELDS:
{example_output_fields}

CONTEXT:
{{{example_context}}}
{{{second_example_context}}}
{{{third_example_context}}}{image_disclaimer}{audio_disclaimer}

AGGREGATION INSTRUCTION: {example_agg_instruction}

Let's think step-by-step in order to answer the question.

REASONING: {example_reasoning}

ANSWER:
{{{example_answer}}}
---
"""

AGG_NO_REASONING_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and an output field to generate. Your task is to generate a JSON object which aggregates the input and fills in the output field with the correct value.
You will be provided with a description of each input field and each output field. The field in the output JSON object can be derived using information from the context.

{output_format_instruction} Finish your response with a newline character followed by ---

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

OUTPUT FIELDS:
{example_output_fields}

CONTEXT:
{{{example_context}}}
{{{second_example_context}}}
{{{third_example_context}}}{image_disclaimer}{audio_disclaimer}

AGGREGATION INSTRUCTION: {example_agg_instruction}

ANSWER:
{{{example_answer}}}
---
"""


AGG_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and an output field to generate. Your task is to generate a JSON object which aggregates the input and fills in the output field with the correct value.
You will be provided with a description of each input field and each output field. The field in the output JSON object can be derived using information from the context.
{desc_section}
{output_format_instruction} Finish your response with a newline character followed by ---
---
INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

CONTEXT:
{context}<<image-audio-placeholder>>

AGGREGATION INSTRUCTION: {agg_instruction}

Let's think step-by-step in order to answer the question.

REASONING: """

AGG_NO_REASONING_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and an output field to generate. Your task is to generate a JSON object which aggregates the input and fills in the output field with the correct value.
You will be provided with a description of each input field and each output field. The field in the output JSON object can be derived using information from the context.
{desc_section}
{output_format_instruction} Finish your response with a newline character followed by ---
---
INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

CONTEXT:
{context}<<image-audio-placeholder>>

AGGREGATION INSTRUCTION: {agg_instruction}

ANSWER: """



from __future__ import annotations

import logging
from copy import deepcopy

from pydantic.fields import FieldInfo

from palimpzest.constants import Model
from palimpzest.core.data.dataset import Dataset
from palimpzest.core.lib.schemas import get_schema_field_names
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_strategy_type import ExecutionStrategyType
from palimpzest.query.operators.logical import (
    ComputeOperator,
    ConvertScan,
    Distinct,
    FilteredScan,
    JoinOp,
    LimitScan,
    Project,
    SearchOperator,
)
from palimpzest.query.optimizer import (
    IMPLEMENTATION_RULES,
    TRANSFORMATION_RULES,
)
from palimpzest.query.optimizer.cost_model import BaseCostModel, SampleBasedCostModel
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.query.optimizer.primitives import Group, LogicalExpression
from palimpzest.query.optimizer.rules import (
    CritiqueAndRefineRule,
    LLMConvertBondedRule,
    MixtureOfAgentsRule,
    RAGRule,
    SplitRule,
)
from palimpzest.query.optimizer.tasks import (
    ApplyRule,
    ExploreGroup,
    OptimizeGroup,
    OptimizeLogicalExpression,
    OptimizePhysicalExpression,
)

logger = logging.getLogger(__name__)


class Optimizer:
    """
    The optimizer is responsible for searching the space of possible physical plans
    for a user's initial (logical) plan and selecting the one which is closest to
    optimizing the user's policy objective.

    This optimizer is modeled after the Cascades framework for top-down query optimization:
    - Thesis describing Cascades implementation (Chapters 1-3):
      https://15721.courses.cs.cmu.edu/spring2023/papers/17-optimizer2/xu-columbia-thesis1998.pdf

    - Andy Pavlo lecture with walkthrough example: https://www.youtube.com/watch?v=PXS49-tFLcI

    - Original Paper: https://www.cse.iitb.ac.in/infolab/Data/Courses/CS632/2015/Papers/Cascades-graefe.pdf

    Notably, this optimization framework has served as the backbone of Microsoft SQL Server, CockroachDB,
    and a few other important DBMS systems.

    NOTE: the optimizer currently assumes that field names are unique across schemas; we do try to enforce
          this by rewriting field names underneath-the-hood to be "{schema_name}.{field_name}", but this still
          does not solve a situation in which -- for example -- a user uses the pz.URL schema twice in the same
          program. In order to address that situation, we will need to augment our renaming scheme.
    """

    def __init__(
        self,
        policy: Policy,
        cost_model: BaseCostModel,
        available_models: list[Model],
        join_parallelism: int = 64,
        reasoning_effort: str | None = "default",
        api_base: str | None = None,
        verbose: bool = False,
        allow_bonded_query: bool = True,
        allow_rag_reduction: bool = False,
        allow_mixtures: bool = True,
        allow_critic: bool = False,
        allow_split_merge: bool = False,
        optimizer_strategy: OptimizationStrategyType = OptimizationStrategyType.PARETO,
        execution_strategy: ExecutionStrategyType = ExecutionStrategyType.PARALLEL,
        use_final_op_quality: bool = False, # TODO: make this func(plan) -> final_quality
        **kwargs,
    ):
        # store the policy
        self.policy = policy

        # store the cost model
        self.cost_model = cost_model

        # mapping from each group id to its Group object
        self.groups = {}

        # mapping from each expression to its Expression object
        self.expressions = {}

        # the stack of tasks to perform during optimization
        self.tasks_stack = []

        # the lists of implementation and transformation rules that the optimizer can apply
        self.implementation_rules = IMPLEMENTATION_RULES
        self.transformation_rules = TRANSFORMATION_RULES

        # get the strategy class associated with the optimizer strategy
        optimizer_strategy_cls = optimizer_strategy.value
        self.strategy = optimizer_strategy_cls()

        # remove transformation rules for optimization strategies which do not require them
        if optimizer_strategy.no_transformation():
            self.transformation_rules = []

        # if we are not performing optimization, set available models to be single model
        # and remove all optimizations (except for bonded queries)
        if optimizer_strategy == OptimizationStrategyType.NONE:
            self.allow_bonded_query = True
            self.allow_rag_reduction = False
            self.allow_mixtures = False
            self.allow_critic = False
            self.allow_split_merge = False
            self.available_models = [available_models[0]]

        # store optimization hyperparameters
        self.verbose = verbose
        self.available_models = available_models
        self.join_parallelism = join_parallelism
        self.reasoning_effort = reasoning_effort
        self.api_base = api_base
        self.allow_bonded_query = allow_bonded_query
        self.allow_rag_reduction = allow_rag_reduction
        self.allow_mixtures = allow_mixtures
        self.allow_critic = allow_critic
        self.allow_split_merge = allow_split_merge
        self.optimizer_strategy = optimizer_strategy
        self.execution_strategy = execution_strategy
        self.use_final_op_quality = use_final_op_quality

        # prune implementation rules based on boolean flags
        if not self.allow_bonded_query:
            self.implementation_rules = [
                rule
                for rule in self.implementation_rules
                if rule not in [LLMConvertBondedRule]
            ]

        if not self.allow_rag_reduction:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, RAGRule)
            ]

        if not self.allow_mixtures:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, MixtureOfAgentsRule)
            ]

        if not self.allow_critic:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, CritiqueAndRefineRule)
            ]

        if not self.allow_split_merge:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, SplitRule)
            ]

        logger.info(f"Initialized Optimizer with verbose={self.verbose}")
        logger.debug(f"Initialized Optimizer with params: {self.__dict__}")

    def update_cost_model(self, cost_model: BaseCostModel):
        self.cost_model = cost_model

    def get_physical_op_params(self):
        return {
            "verbose": self.verbose,
            "available_models": self.available_models,
            "join_parallelism": self.join_parallelism,
            "reasoning_effort": self.reasoning_effort,
            "api_base": self.api_base,
            "is_validation": self.optimizer_strategy == OptimizationStrategyType.SENTINEL,
        }

    def deepcopy_clean(self):
        optimizer = Optimizer(
            policy=self.policy,
            cost_model=SampleBasedCostModel(),
            verbose=self.verbose,
            available_models=self.available_models,
            join_parallelism=self.join_parallelism,
            reasoning_effort=self.reasoning_effort,
            api_base=self.api_base,
            allow_bonded_query=self.allow_bonded_query,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            allow_critic=self.allow_critic,
            allow_split_merge=self.allow_split_merge,
            optimizer_strategy=self.optimizer_strategy,
            execution_strategy=self.execution_strategy,
            use_final_op_quality=self.use_final_op_quality,
        )
        return optimizer

    def update_strategy(self, optimizer_strategy: OptimizationStrategyType):
        # set the optimizer_strategy
        self.optimizer_strategy = optimizer_strategy

        # get the strategy class associated with the optimizer strategy
        optimizer_strategy_cls = optimizer_strategy.value
        self.strategy = optimizer_strategy_cls()

        # remove transformation rules for optimization strategies which do not require them
        if optimizer_strategy.no_transformation():
            self.transformation_rules = []

    def construct_group_tree(self, dataset: Dataset) -> tuple[int, dict[str, FieldInfo], dict[str, set[str]]]:
        logger.debug(f"Constructing group tree for dataset: {dataset}")
        ### convert node --> Group ###
        # create the op for the given node
        op = dataset._operator

        # compute the input group id(s) and field(s) for this node
        if len(dataset._sources) == 0:
            input_group_ids, input_group_fields, input_group_properties = ([], {}, {})
        elif len(dataset._sources) == 1:
            input_group_id, input_group_fields, input_group_properties = self.construct_group_tree(dataset._sources[0])
            input_group_ids = [input_group_id]
        elif len(dataset._sources) == 2:
            left_input_group_id, left_input_group_fields, left_input_group_properties = self.construct_group_tree(dataset._sources[0])
            right_input_group_id, right_input_group_fields, right_input_group_properties = self.construct_group_tree(dataset._sources[1])
            input_group_ids = [left_input_group_id, right_input_group_id]
            input_group_fields = {**left_input_group_fields, **right_input_group_fields}
            input_group_properties = deepcopy(left_input_group_properties)
            for k, v in right_input_group_properties.items():
                if k in input_group_properties:
                    input_group_properties[k].update(v)
                else:
                    input_group_properties[k] = deepcopy(v)
        else:
            raise NotImplementedError("Constructing group trees for datasets with more than 2 sources is not supported.")

        # compute the fields added by this operation and all fields
        input_group_short_field_names = list(
            map(lambda full_field: full_field.split(".")[-1], input_group_fields.keys())
        )
        new_fields = {
            field_name: op.output_schema.model_fields[field_name.split(".")[-1]]
            for field_name in get_schema_field_names(op.output_schema, id=dataset.id)
            if (field_name not in input_group_short_field_names) or (hasattr(op, "udf") and op.udf is not None)
        }
        all_fields = {**input_group_fields, **new_fields}

        # compute the set of (short) field names this operation depends on
        depends_on_field_names = (
            {} if dataset.is_root else {field_name.split(".")[-1] for field_name in op.depends_on}
        )

        # NOTE: group_id is computed as the unique (sorted) set of fields and properties;
        #       If an operation does not modify the fields (or modifies them in a way that
        #       can create an idential field set to an earlier group) then we must add an
        #       id from the operator to disambiguate the two groups.
        # compute all properties including this operations'
        all_properties = deepcopy(input_group_properties)
        if isinstance(op, ConvertScan) and sorted(op.input_schema.model_fields.keys()) == sorted(op.output_schema.model_fields.keys()):
            model_fields_dict = {
                k: {"annotation": v.annotation, "default": v.default, "description": v.description}
                for k, v in op.output_schema.model_fields.items()
            }
            if "maps" in all_properties:
                all_properties["maps"].add(model_fields_dict)
            else:
                all_properties["maps"] = set([model_fields_dict])

        elif isinstance(op, FilteredScan):
            # NOTE: we could use op.get_full_op_id() here, but storing filter strings makes
            #       debugging a bit easier as you can read which filters are in the Group
            op_filter_str = op.filter.get_filter_str()
            if "filters" in all_properties:
                all_properties["filters"].add(op_filter_str)
            else:
                all_properties["filters"] = set([op_filter_str])

        elif isinstance(op, JoinOp):
            unique_join_str = str(sorted(op.on)) if op.condition is None else op.condition
            if "joins" in all_properties:
                all_properties["joins"].add(unique_join_str)
            else:
                all_properties["joins"] = set([unique_join_str])

        elif isinstance(op, LimitScan):
            op_limit_str = op.get_logical_op_id()
            if "limits" in all_properties:
                all_properties["limits"].add(op_limit_str)
            else:
                all_properties["limits"] = set([op_limit_str])

        elif isinstance(op, Project):
            op_project_str = op.get_logical_op_id()
            if "projects" in all_properties:
                all_properties["projects"].add(op_project_str)
            else:
                all_properties["projects"] = set([op_project_str])

        elif isinstance(op, Distinct):
            op_distinct_str = op.get_logical_op_id()
            if "distincts" in all_properties:
                all_properties["distincts"].add(op_distinct_str)
            else:
                all_properties["distincts"] = set([op_distinct_str])

        # TODO: temporary fix; perhaps use op_ids to identify group?
        elif isinstance(op, ComputeOperator):
            op_instruction = op.instruction
            if "instructions" in all_properties:
                all_properties["instructions"].add(op_instruction)
            else:
                all_properties["instructions"] = set([op_instruction])

        elif isinstance(op, SearchOperator):
            op_search_query = op.search_query
            if "search_queries" in all_properties:
                all_properties["search_queries"].add(op_search_query)
            else:
                all_properties["search_queries"] = set([op_search_query])

        # construct the logical expression and group
        logical_expression = LogicalExpression(
            operator=op,
            input_group_ids=input_group_ids,
            input_fields=input_group_fields,
            depends_on_field_names=depends_on_field_names,
            generated_fields=new_fields,
            group_id=None,
        )
        group = Group(
            logical_expressions=[logical_expression],
            fields=all_fields,
            properties=all_properties,
        )
        logical_expression.set_group_id(group.group_id)

        # add the expression and group to the optimizer's expressions and groups and return
        self.expressions[logical_expression.expr_id] = logical_expression
        self.groups[group.group_id] = group
        logger.debug(f"Constructed group tree for dataset: {dataset}")
        logger.debug(f"Group: {group.group_id}, {all_fields}, {all_properties}")

        return group.group_id, all_fields, all_properties

    def convert_query_plan_to_group_tree(self, dataset: Dataset) -> str:
        logger.debug(f"Converting query plan to group tree for dataset: {dataset}")

        # compute depends_on field for every node
        short_to_full_field_name = {}
        for node in dataset:
            # update mapping from short to full field names
            short_field_names = get_schema_field_names(node.schema)
            full_field_names = get_schema_field_names(node.schema, id=node.id)
            for short_field_name, full_field_name in zip(short_field_names, full_field_names):
                # set mapping automatically if this is a new field
                if short_field_name not in short_to_full_field_name or (hasattr(node._operator, "udf") and node._operator.udf is not None):
                    short_to_full_field_name[short_field_name] = full_field_name

            # if the node is a root Dataset, then skip
            if node.is_root:
                continue

            # If the node already has depends_on specified, then resolve each field name to a full (unique) field name
            if len(node._operator.depends_on) > 0:
                node._operator.depends_on = list(map(lambda field: short_to_full_field_name[field], node._operator.depends_on))
                continue

            # otherwise, make the node depend on all upstream nodes
            node._operator.depends_on = set()
            upstream_nodes = node.get_upstream_datasets()
            for upstream_node in upstream_nodes:
                upstream_field_names = get_schema_field_names(upstream_node.schema, id=upstream_node.id)
                node._operator.depends_on.update(upstream_field_names)
            node._operator.depends_on = list(node._operator.depends_on)

        # construct tree of groups
        final_group_id, _, _ = self.construct_group_tree(dataset)

        logger.debug(f"Converted query plan to group tree for dataset: {dataset}")
        logger.debug(f"Final group id: {final_group_id}")

        return final_group_id

    def heuristic_optimization(self, group_id: int) -> None:
        """
        Apply universally desirable transformations (e.g. filter/projection push-down).
        """
        pass

    def search_optimization_space(self, group_id: int) -> None:
        logger.debug(f"Searching optimization space for group_id: {group_id}")

        # begin the search for an optimal plan with a task to optimize the final group
        initial_task = OptimizeGroup(group_id)
        self.tasks_stack.append(initial_task)

        # TODO: conditionally stop when X number of tasks have been executed to limit exhaustive search
        while len(self.tasks_stack) > 0:
            task = self.tasks_stack.pop(-1)

            new_tasks = []
            if isinstance(task, (OptimizeGroup, ExploreGroup)):
                new_tasks = task.perform(self.groups)
            elif isinstance(task, OptimizeLogicalExpression):
                new_tasks = task.perform(self.transformation_rules, self.implementation_rules)
            elif isinstance(task, ApplyRule):
                context = {"costed_full_op_ids": self.cost_model.get_costed_full_op_ids()}
                new_tasks = task.perform(
                    self.groups, self.expressions, context=context, **self.get_physical_op_params(),
                )
            elif isinstance(task, OptimizePhysicalExpression):
                context = {"optimizer_strategy": self.optimizer_strategy, "execution_strategy": self.execution_strategy}
                new_tasks = task.perform(self.cost_model, self.groups, self.policy, context=context)

            self.tasks_stack.extend(new_tasks)

        logger.debug(f"Done searching optimization space for group_id: {group_id}")

    def optimize(self, dataset: Dataset) -> list[PhysicalPlan]:
        """
        The optimize function takes in an initial query plan and searches the space of
        logical and physical plans in order to cost and produce a (near) optimal physical plan.
        """
        logger.info(f"Optimizing query plan: {dataset}")
        # compute the initial group tree for the user plan
        dataset_copy = dataset.copy()
        final_group_id = self.convert_query_plan_to_group_tree(dataset_copy)

        # TODO
        # # do heuristic based pre-optimization
        # self.heuristic_optimization(final_group_id)

        # search the optimization space by applying logical and physical transformations to the initial group tree
        self.search_optimization_space(final_group_id)
        logger.info(f"Getting optimal plans for final group id: {final_group_id}")

        return self.strategy.get_optimal_plans(self.groups, final_group_id, self.policy, self.use_final_op_quality)
